#ifndef POSITIONINGSYSTEM_H
#define POSITIONINGSYSTEM_H

//--------------------------------------------------------------------------------------
// File: PositioningSystem.h
// Author: Peter Cerno <petercerno@gmail.com>
//
// Positioning system header
// CPositioningSystem class captures the movements of the mouse
//
//--------------------------------------------------------------------------------------

#include "MsgObserverPattern.h"
#include "Viewport.h"

//--------------------------------------------------------------------------------------
// CPositioningSystem class implements IMsgObserver interface
//--------------------------------------------------------------------------------------
class CPositioningSystem: public IMsgObserver
{
protected:
	CPositioningSystem(const CPositioningSystem &);
	const CPositioningSystem & operator=(const CPositioningSystem &);
public:
	CPositioningSystem(const CViewport & Viewport):
	  m_Viewport(Viewport),
	  m_globalStartX(0),
	  m_globalStartY(0),
	  m_globalActualX(0),
	  m_globalActualY(0),
	  m_globalFinishX(0),
	  m_globalFinishY(0),
	  m_bStartDefined(false),
	  m_bActualDefined(false),
	  m_bFinishDefined(false),
	  m_localStart(0.0f, 0.0f),
	  m_localActual(0.0f, 0.0f),
	  m_localFinish(0.0f, 0.0f),
	  m_LButtonDown(false),
	  //m_RButtonDown(false),
	  m_bCaptured(false)
	  { }
	virtual ~CPositioningSystem() { }

	void MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

	bool IsStartDefined()  const { return m_bStartDefined; }
	bool IsActualDefined() const { return m_bActualDefined; }
	bool IsFinishDefined() const { return m_bFinishDefined; }

	void Done() { m_bStartDefined = m_bFinishDefined = m_bCaptured = false; }

	int GetGlobalStartX()  const { if (m_bStartDefined)  return m_globalStartX;  else return 0; }
	int GetGlobalStartY()  const { if (m_bStartDefined)  return m_globalStartY;  else return 0; }
	int GetGlobalActualX() const { if (m_bActualDefined) return m_globalActualX; else return 0; }
	int GetGlobalActualY() const { if (m_bActualDefined) return m_globalActualY; else return 0; }
	int GetGlobalFinishX() const { if (m_bFinishDefined) return m_globalFinishX; else return 0; }
	int GetGlobalFinishY() const { if (m_bFinishDefined) return m_globalFinishY; else return 0; }

	float GetLocalStartX()  const { if (m_bStartDefined)  return m_localStart.x;  else return 0.0f; }
	float GetLocalStartY()  const { if (m_bStartDefined)  return m_localStart.y;  else return 0.0f; }
	float GetLocalActualX() const { if (m_bActualDefined) return m_localActual.x; else return 0.0f; }
	float GetLocalActualY() const { if (m_bActualDefined) return m_localActual.y; else return 0.0f; }
	float GetLocalFinishX() const { if (m_bFinishDefined) return m_localFinish.x; else return 0.0f; }
	float GetLocalFinishY() const { if (m_bFinishDefined) return m_localFinish.y; else return 0.0f; }

	D3DXVECTOR2 GetLocalStart()  const { if (m_bStartDefined)  return m_localStart;  else return D3DXVECTOR2(0.0f, 0.0f); } 
	D3DXVECTOR2 GetLocalActual() const { if (m_bActualDefined) return m_localActual; else return D3DXVECTOR2(0.0f, 0.0f); } 
	D3DXVECTOR2 GetLocalFinish() const { if (m_bActualDefined) return m_localFinish; else return D3DXVECTOR2(0.0f, 0.0f); } 

protected:
	const CViewport & m_Viewport;

	int m_globalStartX, m_globalStartY;
	int m_globalActualX, m_globalActualY;
	int m_globalFinishX, m_globalFinishY;

	D3DXVECTOR2 m_localStart;
	D3DXVECTOR2 m_localActual;
	D3DXVECTOR2 m_localFinish;

	bool m_bStartDefined;
	bool m_bActualDefined;
	bool m_bFinishDefined;

	bool m_LButtonDown;
	//bool m_RButtonDown;

	bool m_bCaptured;

	void OnLeftDown(int gX, int gY);
	void OnLeftUp(int gX, int gY);
	void OnMove(int gX, int gY);
};

void CPositioningSystem::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	int x = (short)(LOWORD(lParam));
	int y = (short)(HIWORD(lParam));
	m_LButtonDown = false;
//	m_RButtonDown = false;
	switch(uMsg)
	{
	case WM_LBUTTONDOWN: 
		SetCapture(hWnd);
		OnLeftDown(x, y);
		break;
	case WM_LBUTTONUP:
		OnLeftUp(x, y);
		ReleaseCapture();
		break;
/*	case WM_RBUTTONDOWN:
		SetCapture(hWnd);
		OnRightDown(x, y);
		break;
	case WM_RBUTTONUP:
		OnRightUp(x, y);
		ReleaseCapture();
		break;*/
	case WM_MOUSEMOVE:
		m_LButtonDown = wParam & MK_LBUTTON;
//		m_RButtonDown = wParam & MK_RBUTTON;
		OnMove(x, y);
		break;
	}
}

void CPositioningSystem::OnLeftDown(int gX, int gY)
{
	if (m_Viewport.IsInside(gX, gY))
	{
		m_bCaptured = true;
		m_LButtonDown = true;

		m_globalStartX = gX;
		m_globalStartY = gY;
		m_Viewport.GetLocalPoint(gX, gY, m_localStart);
		m_bStartDefined = true;
	}
}

void CPositioningSystem::OnMove(int gX, int gY)
{
	if (m_bCaptured || m_Viewport.IsInside(gX, gY))
	{
		m_globalActualX = gX;
		m_globalActualY = gY;
		m_Viewport.GetLocalPoint(gX, gY, m_localActual);
		m_bActualDefined = true;
	}
	else
		m_bActualDefined = false;
}

void CPositioningSystem::OnLeftUp(int gX, int gY)
{
	if (m_bCaptured && m_bStartDefined)
	{
		m_globalFinishX = gX;
		m_globalFinishY = gY;
		m_Viewport.GetLocalPoint(gX, gY, m_localFinish);
		m_bFinishDefined = true;
		m_bCaptured = false;
	}
}

#endif // POSITIONINGSYSTEM_H