#ifndef RFWALGEBRA_H
#define RFWALGEBRA_H

//--------------------------------------------------------------------------------------
// File: RFWAlgebra.h
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core classes extension header
// CVector and CMatrix classes
//
//--------------------------------------------------------------------------------------

#include <cassert>
#include <string>
#include <vector>

namespace RFW
{
	class CVector;
	class CMatrix;
	class CVectorReader;
	class CMatrixReader;
}

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CVector class
	//--------------------------------------------------------------------------------------
	class CVector
	{
	public:
		CVector(): m_M(0) { }
		explicit CVector(int m):
		  m_M(m)
		{
			m_Vector.resize(m);
		}
		virtual ~CVector() { }

		void Clear();

		int GetM() const { return m_M; }

		double GetAt(int i) const   { return m_Vector[i]; }
		void SetAt(int i, double x) { m_Vector[i] = x; }

		void Apply(double (*function)(double));

		const CVector & operator+() const { return *this; }
		const CVector operator-() const;

		// WARNING: Unconventional operations * and / (element by element) !!!
		const CVector & operator+=(const CVector & rhs);
		const CVector & operator-=(const CVector & rhs);
		const CVector & operator*=(const CVector & rhs);
		const CVector & operator/=(const CVector & rhs);
		friend const CVector operator+(const CVector & lhs, const CVector & rhs);
		friend const CVector operator-(const CVector & lhs, const CVector & rhs);
		friend const CVector operator*(const CVector & lhs, const CVector & rhs);
		friend const CVector operator/(const CVector & lhs, const CVector & rhs);
		friend const CVector operator*(const CMatrix & lhs, const CVector & rhs);
	protected:
		int m_M;
		std::vector<double> m_Vector;

		friend CVectorReader;
	};

	//--------------------------------------------------------------------------------------
	// CMatrix class
	//--------------------------------------------------------------------------------------
	class CMatrix
	{
	public:
		CMatrix(): m_M(0), m_N(0) { }
		explicit CMatrix(int m, int n):
		  m_M(m), m_N(n)
		{
			m_Matrix.resize(m);
			for (int i = 0; i < m; ++i)
				m_Matrix[i].resize(n);
		}
		virtual ~CMatrix() { }

		void Clear();

		int GetM() const { return m_M; }
		int GetN() const { return m_N; }

		double GetAt(int i, int j) const   { return m_Matrix[i][j]; }
		void SetAt(int i, int j, double x) { m_Matrix[i][j] = x; }

		const CMatrix & operator+() const { return *this; }
		const CMatrix operator-() const;

		const CMatrix & operator+=(const CMatrix & rhs);
		const CMatrix & operator-=(const CMatrix & rhs);
		friend const CMatrix operator+(const CMatrix & lhs, const CMatrix & rhs);
		friend const CMatrix operator-(const CMatrix & lhs, const CMatrix & rhs);
		friend const CMatrix operator*(const CMatrix & lhs, const CMatrix & rhs);
		friend const CVector operator*(const CMatrix & lhs, const CVector & rhs);
	protected:
		int m_M, m_N;
		std::vector< std::vector<double> > m_Matrix;

		friend CMatrixReader;
	};

	inline void CVector::Clear()
	{
		m_Vector.clear();
		m_M = 0;
	}

	inline void CVector::Apply(double (*function)(double))
	{
		for (int i = 0; i < m_M; ++i)
			m_Vector[i] = function(m_Vector[i]);
	}

	inline const CVector CVector::operator-() const
	{
		CVector vector(m_M);
		for (int i = 0; i < m_M; ++i)
			vector.m_Vector[i] = -m_Vector[i];
		return vector;
	}

	inline const CVector & CVector::operator+=(const CVector & rhs)
	{
		assert(m_M == rhs.m_M);
		for (int i = 0; i < m_M; ++i)
			m_Vector[i] += rhs.m_Vector[i];
		return *this;
	}

	inline const CVector & CVector::operator-=(const CVector & rhs)
	{
		assert(m_M == rhs.m_M);
		for (int i = 0; i < m_M; ++i)
			m_Vector[i] -= rhs.m_Vector[i];
		return *this;
	}

	inline const CVector & CVector::operator*=(const CVector & rhs)
	{
		assert(m_M == rhs.m_M);
		for (int i = 0; i < m_M; ++i)
			m_Vector[i] *= rhs.m_Vector[i];
		return *this;
	}

	inline const CVector & CVector::operator/=(const CVector & rhs)
	{
		assert(m_M == rhs.m_M);
		for (int i = 0; i < m_M; ++i)
		{
			if (rhs.m_Vector[i] > -1.0e-10 &&
				rhs.m_Vector[i] < +1.0e-10)
				m_Vector[i] = 0.0;
			else
				m_Vector[i] /= rhs.m_Vector[i];
		}
		return *this;
	}

	inline const CVector operator+(const CVector & lhs, const CVector & rhs)
	{
		CVector vector(lhs);
		vector += rhs;
		return vector;
	}

	inline const CVector operator-(const CVector & lhs, const CVector & rhs)
	{
		CVector vector(lhs);
		vector -= rhs;
		return vector;
	}

	inline const CVector operator*(const CVector & lhs, const CVector & rhs)
	{
		CVector vector(lhs);
		vector *= rhs;
		return vector;
	}

	inline const CVector operator/(const CVector & lhs, const CVector & rhs)
	{
		CVector vector(lhs);
		vector /= rhs;
		return vector;
	}

	inline void CMatrix::Clear()
	{
		for (int i = 0; i < m_M; ++i)
			m_Matrix[i].clear();
		m_Matrix.clear();
		m_M = 0;
		m_N = 0;
	}

	inline const CMatrix CMatrix::operator-() const
	{
		CMatrix matrix(m_M, m_N);
		for (int i = 0; i < m_M; ++i)
			for (int j = 0; j < m_N; ++j)
				matrix.m_Matrix[i][j] = -m_Matrix[i][j];
		return matrix;
	}

	inline const CMatrix & CMatrix::operator+=(const CMatrix & rhs)
	{
		assert(m_M == rhs.m_M && m_N == rhs.m_N);
		for (int i = 0; i < m_M; ++i)
			for (int j = 0; j < m_N; ++j)
				m_Matrix[i][j] += rhs.m_Matrix[i][j];
		return *this;
	}

	inline const CMatrix & CMatrix::operator-=(const CMatrix & rhs)
	{
		assert(m_M == rhs.m_M && m_N == rhs.m_N);
		for (int i = 0; i < m_M; ++i)
			for (int j = 0; j < m_N; ++j)
				m_Matrix[i][j] -= rhs.m_Matrix[i][j];
		return *this;
	}

	inline const CMatrix operator+(const CMatrix & lhs, const CMatrix & rhs)
	{
		CMatrix matrix(lhs);
		matrix += rhs;
		return matrix;
	}

	inline const CMatrix operator-(const CMatrix & lhs, const CMatrix & rhs)
	{
		CMatrix matrix(lhs);
		matrix -= rhs;
		return matrix;
	}

	inline const CMatrix operator*(const CMatrix & lhs, const CMatrix & rhs)
	{
		assert(lhs.m_N == rhs.m_M);
		CMatrix matrix(lhs.m_M, rhs.m_N);
		for (int i = 0; i < lhs.m_M; ++i)
			for (int j = 0; j < rhs.m_N; ++j)
				for (int k = 0; k < lhs.m_N; ++k)
					matrix.m_Matrix[i][j] += lhs.m_Matrix[i][k] * rhs.m_Matrix[k][j];
		return matrix;
	}

	inline const CVector operator*(const CMatrix & lhs, const CVector & rhs)
	{
		assert(lhs.m_N == rhs.m_M);
		CVector vector(lhs.m_M);
		for (int i = 0; i < lhs.m_M; ++i)
			for (int j = 0; j < lhs.m_N; ++j)
				vector.m_Vector[i] += lhs.m_Matrix[i][j] * rhs.m_Vector[j];
		return vector;
	}
}

#endif // RFWALGEBRA_H