//--------------------------------------------------------------------------------------
// File: RFWAlgebraIO.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core IO extension source
// Implementation of CVectorReader and CMatrixReader classes
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)

#include "TextParser.h"

#include "RFWAlgebra.h"
#include "RFWAlgebraIO.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CVectorReader class reads RFW::CVector
	//--------------------------------------------------------------------------------------
	void CVectorReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, CVector & vec)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CVectorReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;
		std::string text;

		int index = 0;

		int M = 0;
		bool readM    = false;
		bool readText = false;

		vec.Clear();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "vector")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "m") {
						if (!readM) {
							M = atoi(attributeValue.c_str());
							if (M > 0) readM = true;
							else { enkiIOStruct.Error("Error in CVectorReader: m <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CVectorReader: double definition of m"); return; }
					}
				}
			}
			for (index = 0; index < listPtr->length; ++index)
			{
				subnodePtr = listPtr->Getitem(index);
				if (subnodePtr->GetnodeType() == MSXML2::NODE_TEXT)
				{
					if (!readText)
					{
						text = subnodePtr->text;
						readText = true;
					}
					else { enkiIOStruct.Error("Error in CVectorReader: double definition of vector data"); return; }
				} else  { enkiIOStruct.Error("Error in CVectorReader: subnodePtr->GetnodeType() != MSXML2::NODE_TEXT"); return; }
			}
			if (readM && readText)
			{
				std::vector<double> Container;
				CTextParser<double> TextParser;
				int Count = TextParser.Parse(Container, text);
				if (TextParser.IsError() || Count != M)
					{ enkiIOStruct.Error("Error in CVectorReader: unable to parse vector data"); return; }
				else
				{
					vec.m_M      = M;
					vec.m_Vector = Container;
				}
			} else { enkiIOStruct.Error("Error in CVectorReader: incomplete definition of vector"); return; }
		} else { enkiIOStruct.Error("Error in CVectorReader: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CMatrixReader class reads RFW::CMatrix
	//--------------------------------------------------------------------------------------
	void CMatrixReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, CMatrix & mat)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CMatrixReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;
		std::string text;

		int index = 0;

		int M = 0;
		int N = 0;
		bool readM    = false;
		bool readN    = false;
		bool readText = false;

		int i = 0;
		int j = 0;
		int k = 0;

		mat.Clear();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "matrix")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "m") {
						if (!readM) {
							M = atoi(attributeValue.c_str());
							if (M > 0) readM = true;
							else { enkiIOStruct.Error("Error in CMatrixReader: m <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CMatrixReader: double definition of m"); return; }
					}
					else if (attributeName == "n") {
						if (!readN) {
							N = atoi(attributeValue.c_str());
							if (N > 0) readN = true;
							else { enkiIOStruct.Error("Error in CMatrixReader: n <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CMatrixReader: double definition of n"); return; }
					}
				}
			}
			for (index = 0; index < listPtr->length; ++index)
			{
				subnodePtr = listPtr->Getitem(index);
				if (subnodePtr->GetnodeType() == MSXML2::NODE_TEXT)
				{
					if (!readText) {
						text = subnodePtr->text;
						readText = true;
					} else { enkiIOStruct.Error("Error in CMatrixReader: double definition of matrix data"); return; }
				} else { enkiIOStruct.Error("Error in CMatrixReader: subnodePtr->GetnodeType() != MSXML2::NODE_TEXT"); return; }
			}
			if (readM && readN && readText)
			{
				std::vector<double> Container;
				CTextParser<double> TextParser;
				int Count = TextParser.Parse(Container, text);
				if (TextParser.IsError() || Count != M * N)
					{ enkiIOStruct.Error("Error in CMatrixReader: unable to parse matrix data"); return; }
				else
				{
					mat.m_M      = M;
					mat.m_N      = N;
					mat.m_Matrix.resize(M);
					k = 0;
					for (i = 0; i < M; ++i)
					{
						mat.m_Matrix[i].resize(N);
						for (j = 0; j < N; ++j)
							mat.m_Matrix[i][j] = Container[k++];
					}
				}
			} else { enkiIOStruct.Error("Error in CMatrixReader: incomplete definition of matrix"); return; }
		} else { enkiIOStruct.Error("Error in CMatrixReader: bad nodeName"); return; }
	}
}