//--------------------------------------------------------------------------------------
// File: RFWCoreClasses.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core classes source
// Implementation of classes for IWorld, IPhysicalObject, IRobot based on Enki
//
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning( disable : 4530 )

#include <algorithm>
#include <vector>

#include "enki/PhysicalEngine.h"

#include "RFWCoreClasses.h"
#include "RFWCoreGraphicsClasses.h"
#include "RFWCoreIOClasses.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CEnkiWorld class implements IWorld interface
	//--------------------------------------------------------------------------------------
	const double CEnkiWorld::defaultWidth  = 100.0;
	const double CEnkiWorld::defaultHeight = 100.0;

	const double timeMin = 0.001;
	const double timeMax = 1.000;

	CEnkiWorld::CEnkiWorld(CVertexBuffer & vertexBuffer):
	  m_VertexBuffer(vertexBuffer)
	{
		ClearWorld();
		m_EnkiWorldGraphics = new CEnkiWorldGraphics(*this);
		m_EnkiWorldIO = new CEnkiWorldIO(*this);
	}

	CEnkiWorld::~CEnkiWorld()
	{
		ClearWorld();
		delete m_EnkiWorldGraphics;
		delete m_EnkiWorldIO;
	}

	void CEnkiWorld::AddPhysicalObject(RFW::IEnkiPhysicalObject & enkiPhysicalObject)
	{
		m_EnkiPhysicalObjects.push_back(&enkiPhysicalObject);
		GetEnkiWorld()->addObject(enkiPhysicalObject.GetEnkiPhysicalObject());
	}

	void CEnkiWorld::AddRobot(RFW::IEnkiRobot & enkiRobot)
	{
		m_EnkiRobots.push_back(&enkiRobot);
		GetEnkiWorld()->addObject(enkiRobot.GetEnkiRobot());
	}

	void CEnkiWorld::ClearWorld(double width, double height)
	{
		m_StartingPoints.clear();

		GetEnkiWorld()->objects.clear();
		GetEnkiWorld()->useWalls = true;
		GetEnkiWorld()->w = width;
		GetEnkiWorld()->h = height;

		size_t objects_count = m_EnkiPhysicalObjects.size();
		size_t robots_count = m_EnkiRobots.size();
		size_t index = 0;

		for (index = 0; index < objects_count; ++index)
			delete m_EnkiPhysicalObjects[index];
		m_EnkiPhysicalObjects.clear();
		for (index = 0; index < robots_count; ++index)
			delete m_EnkiRobots[index];
		m_EnkiRobots.clear();
	}

	void CEnkiWorld::Render()
	{
		m_EnkiWorldGraphics->Render();
	}

	void CEnkiWorld::Read(RFW::CEnkiIOStruct & enkiIOStruct)
	{
		m_EnkiWorldIO->Read(enkiIOStruct);
	}

	void CEnkiWorld::RemovePhysicalObject(RFW::IEnkiPhysicalObject *physicalObject)
	{
		std::vector<IEnkiPhysicalObject *>::iterator it =
			std::find(
				m_EnkiPhysicalObjects.begin(),
				m_EnkiPhysicalObjects.end(),
				physicalObject);
		if (it != m_EnkiPhysicalObjects.end())
		{
			GetEnkiWorld()->removeObject(physicalObject->GetEnkiPhysicalObject());
			m_EnkiPhysicalObjects.erase(it);
			delete physicalObject;
		}
	}

	void CEnkiWorld::RemoveRobot(RFW::IEnkiRobot *robot)
	{
		std::vector<IEnkiRobot *>::iterator it =
			std::find(
				m_EnkiRobots.begin(),
				m_EnkiRobots.end(),
				robot);
		if (it != m_EnkiRobots.end())
		{
			GetEnkiWorld()->removeObject(robot->GetEnkiRobot());
			m_EnkiRobots.erase(it);
			delete robot;
		}

	}

	void CEnkiWorld::Step(double time)
	{
		if (time > timeMax)
			time = timeMax;
		else if (time < timeMin)
			time = timeMin;
		GetEnkiWorld()->step(time);
	}

	//--------------------------------------------------------------------------------------
	// CEnkiPhysicalObject class implements IPhysicalObject interface
	//--------------------------------------------------------------------------------------
	CEnkiPhysicalObject::CEnkiPhysicalObject(RFW::CEnkiWorld & enkiWorld):
		m_VertexBuffer(enkiWorld.m_VertexBuffer),
		m_EnkiWorld(enkiWorld)
	{
		ClearPhysicalObject();
		m_EnkiPhysicalObjectGraphics = new CEnkiPhysicalObjectGraphics(*this);
		m_EnkiPhysicalObjectIO = new CEnkiPhysicalObjectIO(*this);
		enkiWorld.AddPhysicalObject(*this);
	}

	CEnkiPhysicalObject::~CEnkiPhysicalObject()
	{
		ClearPhysicalObject();
		delete m_EnkiPhysicalObjectGraphics;
		delete m_EnkiPhysicalObjectIO;
	}

	void CEnkiPhysicalObject::ClearPhysicalObject()
	{

	}

	void CEnkiPhysicalObject::Render()
	{
		m_EnkiPhysicalObjectGraphics->Render();
	}

	void CEnkiPhysicalObject::Read(RFW::CEnkiIOStruct & enkiIOStruct)
	{
		m_EnkiPhysicalObjectIO->Read(enkiIOStruct);
	}

	//--------------------------------------------------------------------------------------
	// CEnkiRobot class implements IRobot interface
	//--------------------------------------------------------------------------------------
	CEnkiRobot::CEnkiRobot(RFW::CEnkiWorld & enkiWorld):
		m_VertexBuffer(enkiWorld.m_VertexBuffer),
		m_EnkiWorld(enkiWorld)
	{
		ClearRobot();
		m_EnkiRobotGraphics = new CEnkiRobotGraphics(*this);
		m_EnkiRobotIO = new CEnkiRobotIO(*this);
		enkiWorld.AddRobot(*this); 
	}

	CEnkiRobot::~CEnkiRobot()
	{
		ClearRobot();
		delete m_EnkiRobotGraphics;
		delete m_EnkiRobotIO;
	}

	void CEnkiRobot::ClearRobot()
	{

	}

	void CEnkiRobot::Render()
	{
		m_EnkiRobotGraphics->Render();
	}

	void CEnkiRobot::Read(RFW::CEnkiIOStruct & enkiIOStruct)
	{
		m_EnkiRobotIO->Read(enkiIOStruct);
	}
}