#ifndef RFWCORECLASSES_H
#define RFWCORECLASSES_H

//--------------------------------------------------------------------------------------
// File: RFWCoreClasses.h
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core classes header
// Classes for IWorld, IPhysicalObject, IRobot based on Enki
//
//--------------------------------------------------------------------------------------

#include "RFWCore.h"

#include <vector>

namespace Enki
{
	class World;
	class PhysicalObject;
	class Robot;
}

class CVertexBuffer;

namespace RFW
{
	class IEnkiWorld;
	class IEnkiPhysicalObject;
	class IEnkiRobot;

	class CEnkiWorld;
	class CEnkiPhysicalObject;
	class CEnkiRobot;

	class IEnkiGraphics;

	class CEnkiIOStruct;
	class IEnkiIO;
}

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// IEnkiWorld interface extends IWorld interface
	//--------------------------------------------------------------------------------------
	class IEnkiWorld: public IWorld
	{
	protected:
		IEnkiWorld(const IEnkiWorld &);
		const IEnkiWorld & operator=(const IEnkiWorld &);
	public:
		IEnkiWorld() { }
		virtual ~IEnkiWorld() { }

		virtual void AddPhysicalObject(IEnkiPhysicalObject &) = 0;
		virtual void AddRobot(IEnkiRobot &) = 0;

		virtual void RemovePhysicalObject(RFW::IEnkiPhysicalObject *) = 0;
		virtual void RemoveRobot(RFW::IEnkiRobot *) = 0;

		virtual void ClearWorld(double, double) = 0;

		virtual void AddStartingPoint(An::Point point) = 0;
		virtual size_t GetStartingPointCount() = 0;
		virtual An::Point GetStartingPoint(size_t index) = 0;

		virtual Enki::World * GetEnkiWorld() = 0;

		virtual void Read(CEnkiIOStruct &) = 0;

		virtual void Render() = 0;
	};

	//--------------------------------------------------------------------------------------
	// IEnkiPhysicalObject interface extends IPhysicalObject interface
	//--------------------------------------------------------------------------------------
	class IEnkiPhysicalObject: public IPhysicalObject
	{
	protected:
		IEnkiPhysicalObject(const IEnkiPhysicalObject &);
		const IEnkiPhysicalObject & operator=(const IEnkiPhysicalObject &);
	public:
		explicit IEnkiPhysicalObject() { }
		virtual ~IEnkiPhysicalObject() { }

		virtual void ClearPhysicalObject() = 0;

		virtual Enki::PhysicalObject * GetEnkiPhysicalObject() = 0;

		virtual void Read(CEnkiIOStruct &) = 0;

		virtual void Render() = 0;
	};

	//--------------------------------------------------------------------------------------
	// IEnkiRobot interface extends IRobot interface
	//--------------------------------------------------------------------------------------
	class IEnkiRobot: public IRobot
	{
	protected:
		IEnkiRobot(const IEnkiRobot &);
		const IEnkiRobot & operator=(const IEnkiRobot &);
	public:
		explicit IEnkiRobot() { }
		virtual ~IEnkiRobot() { }

		virtual void ClearRobot() = 0;

		virtual Enki::Robot * GetEnkiRobot() = 0;

		virtual void Read(CEnkiIOStruct &) = 0;

		virtual void Render() = 0;
	};

	//--------------------------------------------------------------------------------------
	// CEnkiWorld class implements IEnkiWorld interface
	// WARNING: This class should not be subclassed
	//--------------------------------------------------------------------------------------
	class CEnkiWorld: public IEnkiWorld
	{
	protected:
		CEnkiWorld(const CEnkiWorld &);
		const CEnkiWorld & operator=(const CEnkiWorld &);
	public:
		static const double defaultWidth;
		static const double defaultHeight;
	public:
		explicit CEnkiWorld(CVertexBuffer &);
		virtual ~CEnkiWorld();

		virtual void AddPhysicalObject(IEnkiPhysicalObject &);
		virtual void AddRobot(IEnkiRobot &);

		virtual void Render();

		virtual void ClearWorld(
			double width = defaultWidth,
			double height = defaultHeight);

		virtual void Read(CEnkiIOStruct &);

		virtual Enki::World * GetEnkiWorld() { return &m_World; }

		virtual size_t GetPhysicalObjectCount() { return m_EnkiPhysicalObjects.size(); }
		virtual IEnkiPhysicalObject * GetPhysicalObject(size_t index) { return m_EnkiPhysicalObjects[index]; }
		virtual void RemovePhysicalObject(RFW::IEnkiPhysicalObject *);

		virtual size_t GetRobotCount() { return m_EnkiRobots.size(); }
		virtual IEnkiRobot * GetRobot(size_t index) { return m_EnkiRobots[index]; }
		virtual void RemoveRobot(RFW::IEnkiRobot *);

		virtual void AddStartingPoint(An::Point point) { m_StartingPoints.push_back(point); }
		virtual size_t GetStartingPointCount() { return m_StartingPoints.size(); }
		virtual An::Point GetStartingPoint(size_t index) { return m_StartingPoints[index]; }

		virtual void Step(double);
	public:
		CVertexBuffer & m_VertexBuffer;
	protected:
		Enki::World m_World;
		IEnkiGraphics * m_EnkiWorldGraphics;
		IEnkiIO * m_EnkiWorldIO;
		std::vector<IEnkiPhysicalObject *> m_EnkiPhysicalObjects;
		std::vector<IEnkiRobot *> m_EnkiRobots;
		std::vector<An::Point> m_StartingPoints;
	};

	//--------------------------------------------------------------------------------------
	// CEnkiPhysicalObject class implements IEnkiPhysicalObject interface
	// WARNING: This class should not be subclassed
	//--------------------------------------------------------------------------------------
	class CEnkiPhysicalObject: public IEnkiPhysicalObject
	{
	protected:
		CEnkiPhysicalObject(const CEnkiPhysicalObject &);
		const CEnkiPhysicalObject & operator=(const CEnkiPhysicalObject &);
	public:
		explicit CEnkiPhysicalObject(CEnkiWorld &);
		virtual ~CEnkiPhysicalObject();

		void ClearPhysicalObject();

		virtual void Render();

		virtual void Read(CEnkiIOStruct &);

		virtual Enki::PhysicalObject * GetEnkiPhysicalObject() { return &m_PhysicalObject; }
	public:
		CVertexBuffer & m_VertexBuffer;
		CEnkiWorld & m_EnkiWorld;
	protected:
		Enki::PhysicalObject m_PhysicalObject;
		IEnkiGraphics * m_EnkiPhysicalObjectGraphics;
		IEnkiIO * m_EnkiPhysicalObjectIO;
	};

	//--------------------------------------------------------------------------------------
	// CEnkiRobot class implements IEnkiRobot interface
	// WARNING: This is only a demonstration class and it should not be used
	//   It only shows you how to construct your own classes for concrete
	//   robots like CEnkiAliceRobot or CEnkiKheperaRobot
	// WARNING: This class should not be subclassed
	//--------------------------------------------------------------------------------------
	class CEnkiRobot: public IEnkiRobot
	{
	protected:
		CEnkiRobot(const CEnkiRobot &);
		const CEnkiRobot & operator=(const CEnkiRobot &);
	public:
		explicit CEnkiRobot(CEnkiWorld &);
		virtual ~CEnkiRobot();

		void ClearRobot();

		virtual void Render();

		virtual void Read(CEnkiIOStruct &);

		virtual Enki::Robot * GetEnkiRobot() { return &m_Robot; }
	public:
		CVertexBuffer & m_VertexBuffer;
		CEnkiWorld & m_EnkiWorld;
	protected:
		Enki::Robot m_Robot;
		IEnkiGraphics * m_EnkiRobotGraphics;
		IEnkiIO * m_EnkiRobotIO;
	};
}

#endif // RFWCORECLASSES_H