//--------------------------------------------------------------------------------------
// File: RFWCoreGraphicsClasses.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core graphics source
// Implementation of graphics classes for IWorld, IPhysicalObject, IRobot based on Enki
//
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning(disable : 4530)

#include "enki/PhysicalEngine.h"
#include "enki/Interaction.h"
#include "enki/interactions/CircularCam.h"
#include "enki/interactions/IRSensor.h"

#include "CustomVertex.h"
#include "VertexBuffer.h"

#include "RFWCoreClasses.h"
#include "RFWCoreGraphicsClasses.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CEnkiPhysicalObjectPlotter class plots Enki::PhysicalObject
	//--------------------------------------------------------------------------------------
	HRESULT CEnkiPhysicalObjectPlotter::Render(Enki::PhysicalObject &physicalObject)
	{
		HRESULT hr;

		CUSTOMVERTEX * vertices = 0;
		size_t Count = 0;
		size_t Index = 0;

		V_RETURN( m_VertexBuffer.Clear() );

		DWORD Color = D3DCOLOR_COLORVALUE(
			physicalObject.color.components[0],
			physicalObject.color.components[1],
			physicalObject.color.components[2],
			physicalObject.color.components[3]);

		if (physicalObject.boundingSurface)
		{
			Count = physicalObject.absBoundingSurface.size();
			vertices = new CUSTOMVERTEX[Count + 2];
			vertices[0].x = (float)physicalObject.pos.x;
			vertices[0].y = (float)physicalObject.pos.y;
			vertices[0].z = 0.5f;
			vertices[0].color = Color;
			for (Index = 1; Index <= Count; ++Index)
			{
				vertices[Index].x = (float)physicalObject.absBoundingSurface[Index - 1].x;
				vertices[Index].y = (float)physicalObject.absBoundingSurface[Index - 1].y;
				vertices[Index].z = 0.5f;
				vertices[Index].color = Color;
			}
			vertices[Count + 1].x = (float)physicalObject.absBoundingSurface[0].x;
			vertices[Count + 1].y = (float)physicalObject.absBoundingSurface[0].y;
			vertices[Count + 1].z = 0.5f;
			vertices[Count + 1].color = Color;
		}
		else
		{
			Count = 32;
			vertices = new CUSTOMVERTEX[Count + 2];
			vertices[0].x = (float)physicalObject.pos.x;
			vertices[0].y = (float)physicalObject.pos.y;
			vertices[0].z = 0.5f;
			vertices[0].color = Color;
			for (Index = 1; Index <= Count + 1; ++Index)
			{
				double Angle = 2 * M_PI * (double)(Index - 1) / (double)Count + physicalObject.angle;
				vertices[Index].x = (float)(physicalObject.pos.x + physicalObject.r * cos(Angle));
				vertices[Index].y = (float)(physicalObject.pos.y + physicalObject.r * sin(Angle));
				vertices[Index].z = 0.5f;
				vertices[Index].color = Color;
			}
		}

		V_RETURN( m_VertexBuffer.CopyVertices(vertices, ((UINT)Count + 2) * sizeof(CUSTOMVERTEX)) );

		delete[] vertices;

		V_RETURN( m_VertexBuffer.SetVertexBuffer() );
		V_RETURN( m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLEFAN) );

		return S_OK;
	}

	//--------------------------------------------------------------------------------------
	// CEnkiCircularCamPlotter class plots Enki::CircularCam
	//--------------------------------------------------------------------------------------
	HRESULT CEnkiCircularCamPlotter::Render(Enki::CircularCam &circularCam, double r)
	{
		HRESULT hr;

		CUSTOMVERTEX * vertices = 0;
		size_t Count = circularCam.image.size();
		size_t i = 0;

		V_RETURN( m_VertexBuffer.Clear() );

		vertices = new CUSTOMVERTEX[2 * Count];

		for (i = 0; i < Count; ++i)
		{
			double x0 = circularCam.getAbsolutePosition().x;
			double y0 = circularCam.getAbsolutePosition().y;
			double Angle = circularCam.getAbsoluteOrientation() -
				circularCam.fieldOfView +
				2.0 * circularCam.fieldOfView * (double)i / (double)(Count - 1);
			double r1 = 1.5 * r;
			double r2 = 2.5 * r;
			double x1 = x0 + cos(Angle) * r1;
			double y1 = y0 + sin(Angle) * r1;
			double x2 = x0 + cos(Angle) * r2;
			double y2 = y0 + sin(Angle) * r2;
			vertices[2*i + 0].x = (float)x1;
			vertices[2*i + 0].y = (float)y1;
			vertices[2*i + 0].z = 0.4f;
			vertices[2*i + 0].color = D3DCOLOR_COLORVALUE(
				circularCam.image[i].components[0],
				circularCam.image[i].components[1],
				circularCam.image[i].components[2],
				circularCam.image[i].components[3]);
			vertices[2*i + 1].x = (float)x2;
			vertices[2*i + 1].y = (float)y2;
			vertices[2*i + 1].z = 0.4f;
			vertices[2*i + 1].color = D3DCOLOR_COLORVALUE(
				circularCam.image[i].components[0],
				circularCam.image[i].components[1],
				circularCam.image[i].components[2],
				circularCam.image[i].components[3]);
		}

		V_RETURN( m_VertexBuffer.CopyVertices(vertices, 2 * (UINT)Count * sizeof(CUSTOMVERTEX)) );

		delete[] vertices;

		V_RETURN( m_VertexBuffer.SetVertexBuffer() );
		V_RETURN( m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLESTRIP) );

		return S_OK;
	}

	//--------------------------------------------------------------------------------------
	// CEnkiIRSensorPlotter class plots Enki::IRSensor
	//--------------------------------------------------------------------------------------
	HRESULT CEnkiIRSensorPlotter::Render(Enki::IRSensor &irSensor)
	{
		HRESULT hr;

		CUSTOMVERTEX * vertices = 0;
		size_t Count = irSensor.getRayCount();
		size_t i = 0;

		V_RETURN( m_VertexBuffer.Clear() );

		vertices = new CUSTOMVERTEX[2 * Count];

		for (i = 0; i < Count; ++i)
		{
			vertices[2*i + 0].x = (float)irSensor.getAbsolutePosition().x;
			vertices[2*i + 0].y = (float)irSensor.getAbsolutePosition().y;
			vertices[2*i + 0].z = 0.8f;
			vertices[2*i + 0].color = D3DCOLOR_COLORVALUE(
				irSensor.finalRayColors[i].components[0],
				irSensor.finalRayColors[i].components[1],
				irSensor.finalRayColors[i].components[2],
				irSensor.finalRayColors[i].components[3]);
			vertices[2*i + 1].x = (float)
				(irSensor.getAbsolutePosition().x + 
				 irSensor.finalRayVectors[i].x);
			vertices[2*i + 1].y = (float)
				(irSensor.getAbsolutePosition().y + 
				 irSensor.finalRayVectors[i].y);
			vertices[2*i + 1].z = 0.8f;
			vertices[2*i + 1].color = D3DCOLOR_COLORVALUE(
				irSensor.finalRayColors[i].components[0],
				irSensor.finalRayColors[i].components[1],
				irSensor.finalRayColors[i].components[2],
				irSensor.finalRayColors[i].components[3]);
		}

		V_RETURN( m_VertexBuffer.AddVertices(vertices, 2 * (UINT)Count * sizeof(CUSTOMVERTEX)) );

		delete[] vertices;

		V_RETURN( m_VertexBuffer.SetVertexBuffer() );
		V_RETURN( m_VertexBuffer.DrawPrimitive(D3DPT_LINELIST) );

		return S_OK;
	}

	//--------------------------------------------------------------------------------------
	// CEnkiWorldGraphics class implements IGraphics interface for CEnkiWorld
	//--------------------------------------------------------------------------------------
	CEnkiWorldGraphics::CEnkiWorldGraphics(RFW::CEnkiWorld &enkiWorld):
		m_VertexBuffer(enkiWorld.m_VertexBuffer),
		m_EnkiWorld(enkiWorld)
	{
		
	}
	
	void CEnkiWorldGraphics::Render()
	{
		float width = (float)m_EnkiWorld.GetEnkiWorld()->w;
		float height = (float)m_EnkiWorld.GetEnkiWorld()->h;
		CUSTOMVERTEX vertices[5] =
		{
			{   0.0f,  0.0f,   0.5f, 0xff000000, }, // x, y, z, color
			{   0.0f,  height, 0.5f, 0xff000000, },
			{   width, height, 0.5f, 0xff000000, },
			{   width, 0.0f,   0.5f, 0xff000000, },
			{   0.0f,  0.0f,   0.5f, 0xff000000, },
		};

		m_VertexBuffer.CopyVertices(vertices, sizeof(vertices));
		m_VertexBuffer.SetVertexBuffer();
		m_VertexBuffer.DrawPrimitive(D3DPT_LINESTRIP);

		size_t po_count = m_EnkiWorld.GetPhysicalObjectCount(); 
		size_t rs_count = m_EnkiWorld.GetRobotCount();
		size_t i = 0;

		for (i = 0; i < po_count; ++i)
			m_EnkiWorld.GetPhysicalObject(i)->Render();
		for (i = 0; i < rs_count; ++i)
			m_EnkiWorld.GetRobot(i)->Render();
	}


	//-----------------------------------------------------------------------------------------
	// CEnkiPhysicalObjectGraphics class implements IGraphics interface for CEnkiPhysicalObject
	//-----------------------------------------------------------------------------------------
	CEnkiPhysicalObjectGraphics::CEnkiPhysicalObjectGraphics(RFW::CEnkiPhysicalObject &enkiPhysicalObject):
		m_VertexBuffer(enkiPhysicalObject.m_VertexBuffer),
		m_EnkiPhysicalObject(enkiPhysicalObject),
		m_EnkiPhysicalObjectPlotter(enkiPhysicalObject.m_VertexBuffer)
	{
	
	}

	void CEnkiPhysicalObjectGraphics::Render()
	{
		m_EnkiPhysicalObjectPlotter.Render(*m_EnkiPhysicalObject.GetEnkiPhysicalObject());
	}

	//--------------------------------------------------------------------------------------
	// CEnkiRobotGraphics class implements IGraphics interface for CEnkiRobot
	//--------------------------------------------------------------------------------------
	CEnkiRobotGraphics::CEnkiRobotGraphics(RFW::CEnkiRobot &enkiRobot):
		m_VertexBuffer(enkiRobot.m_VertexBuffer),
		m_EnkiRobot(enkiRobot),
		m_EnkiPhysicalObjectPlotter(enkiRobot.m_VertexBuffer)
	{
	
	}

	void CEnkiRobotGraphics::Render()
	{
		float width  = (float)m_EnkiRobot.GetEnkiRobot()->r / 2.0f;
		float height = (float)m_EnkiRobot.GetEnkiRobot()->r / 2.0f;
		float angle  = (float)m_EnkiRobot.GetEnkiRobot()->angle;

		CUSTOMVERTEX vertices[3] = 
		{
			{-height, -width, 0.4f, 0xffaaaaaa, }, // x, y, z, color
			{-height, +width, 0.4f, 0xffaaaaaa, },
			{+height,   0.0f, 0.4f, 0xffaaaaaa, },
		};

		m_VertexBuffer.CopyVertices(vertices, 3 * sizeof(CUSTOMVERTEX));
		m_VertexBuffer.SetVertexBuffer();

		D3DXMATRIX matTranslate;
		D3DXMATRIX matRotate;
		D3DXMATRIX matTransform;

		D3DXMatrixTranslation(&matTranslate,
			(float)m_EnkiRobot.GetEnkiRobot()->pos.x,
			(float)m_EnkiRobot.GetEnkiRobot()->pos.y, 0.0f);
		D3DXMatrixRotationZ  (&matRotate, angle);
		D3DXMatrixMultiply(&matTransform, &matRotate, &matTranslate);

		m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLELIST, matTransform);

		m_EnkiPhysicalObjectPlotter.Render(*m_EnkiRobot.GetEnkiRobot());
	}
}