//--------------------------------------------------------------------------------------
// File: RFWCoreIO.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core IO source
// Implementation of IO classes for IWorld, IPhysicalObject, IRobot based on Enki
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)

#include "enki/PhysicalEngine.h"
#include "enki/Interaction.h"
#include "enki/interactions/CircularCam.h"
#include "enki/interactions/IRSensor.h"

#include "TextParser.h"

#include "RFWCoreClasses.h"
#include "RFWCoreIOClasses.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// Color conversion
	//--------------------------------------------------------------------------------------
	An::Color strtocolor(const std::string & colorstr)
	{
		std::string str;
		std::string hexstr;
		{
			size_t strlen = colorstr.length();
			for (size_t i = 0; i < strlen; ++i)
				str.push_back((char)tolower(colorstr[i]));
		}
		if      (str == "aqua") hexstr = "00FFFF";
		else if (str == "azure") hexstr = "F0FFFF";
		else if (str == "black") hexstr = "000000";
		else if (str == "blue") hexstr = "0000FF";
		else if (str == "brown") hexstr = "A52A2A";
		else if (str == "chocolate") hexstr = "D2691E";
		else if (str == "cyan") hexstr = "00FFFF";
		else if (str == "darkblue") hexstr = "00008B";
		else if (str == "darkcyan") hexstr = "008B8B";
		else if (str == "darkgray") hexstr = "A9A9A9";
		else if (str == "darkgrey") hexstr = "A9A9A9";
		else if (str == "darkgreen") hexstr = "006400";
		else if (str == "darkmagenta") hexstr = "8B008B";
		else if (str == "darkolivegreen") hexstr = "556B26";
		else if (str == "darkorange") hexstr = "FF8C00";
		else if (str == "darkred") hexstr = "8B0000";
		else if (str == "darkturquoise") hexstr = "00CED1";
		else if (str == "darkviolet") hexstr = "9400D3";
		else if (str == "gold") hexstr = "FFD700";
		else if (str == "gray") hexstr = "808080";
		else if (str == "grey") hexstr = "808080";
		else if (str == "green") hexstr = "008000";
		else if (str == "greenyellow") hexstr = "ADFF2F";
		else if (str == "indigo") hexstr = "4B0082";
		else if (str == "lightblue") hexstr = "ADD8E6";
		else if (str == "lightcyan") hexstr = "E0FFFF";
		else if (str == "lightgray") hexstr = "D3D3D3";
		else if (str == "lightgrey") hexstr = "D3D3D3";
		else if (str == "lightgreen") hexstr = "90EE90";
		else if (str == "lightpink") hexstr = "FFB6C1";
		else if (str == "lightyellow") hexstr = "FFFFE0";
		else if (str == "lime") hexstr = "00FF00";
		else if (str == "limegreen") hexstr = "32CD32";
		else if (str == "magenta") hexstr = "FF00FF";
		else if (str == "maroon") hexstr = "800000";
		else if (str == "navy") hexstr = "000080";
		else if (str == "olive") hexstr = "808000";
		else if (str == "orange") hexstr = "FFA500";
		else if (str == "orchid") hexstr = "DA70D6";
		else if (str == "pink") hexstr = "FFC0CB";
		else if (str == "purple") hexstr = "800080";
		else if (str == "red") hexstr = "FF0000";
		else if (str == "salmon") hexstr = "FA8072";
		else if (str == "sienna") hexstr = "A0522D";
		else if (str == "silver") hexstr = "C0C0C0";
		else if (str == "turquoise") hexstr = "40E0D0";
		else if (str == "violet") hexstr = "EE82EE";
		else if (str == "white") hexstr = "FFFFFF";
		else if (str == "yellow") hexstr = "FFFF00";
		else if (str == "yellowgreen") hexstr = "9ACD32";
		else
		{
			if (str.length() == 7 && str[0] == '#')
			{
				bool bSuccess = true;
				for (size_t i = 1; i <= 6; ++i)
					if (!((str[i] >= '0' && str[i] <= '9') || 
						  (str[i] >= 'A' && str[i] <= 'F')))
					{
						bSuccess = false;
						break;
					}
				if (bSuccess)
					hexstr = str.substr(1, 6);
				else return An::Color::black;
			}
			else return An::Color::black;
		}
		unsigned int rgb[] = {0, 0, 0};
		unsigned int digit = 0;
		for (size_t i = 0; i < 6; ++i)
		{
			if (hexstr[i] >= '0' && hexstr[i] <= '9')
				digit = hexstr[i] - '0';
			else if (hexstr[i] >= 'A' && hexstr[i] <= 'F')
				digit = hexstr[i] - 'A' + 10;
			rgb[i / 2] = 16 * rgb[i / 2] + digit;
		}
		return An::Color(
			(double)rgb[0] / 255.0,
			(double)rgb[1] / 255.0,
			(double)rgb[2] / 255.0);
	}

	//--------------------------------------------------------------------------------------
	// CAnPolygoneReader class reads An::Polygone
	//--------------------------------------------------------------------------------------
	void CAnPolygoneReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, An::Polygone &polygone)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CAnPolygoneReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;
		std::string text;

		int index = 0;

		int M = 0;
		bool readM    = false;
		bool readText = false;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "polygone")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "m") {
						if (!readM) {
							M = atoi(attributeValue.c_str());
							if (M > 0) readM = true;
							else { enkiIOStruct.Error("Error in CAnPolygoneReader: m <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CAnPolygoneReader: double definition of m"); return; }
					}
				}
			}
			for (index = 0; index < listPtr->length; ++index)
			{
				subnodePtr = listPtr->Getitem(index);
				if (subnodePtr->GetnodeType() == MSXML2::NODE_TEXT)
				{
					if (!readText) {
						text = subnodePtr->text;
						readText = true;
					} else { enkiIOStruct.Error("Error in CAnPolygoneReader: double definition of text"); return; }
				} else { enkiIOStruct.Error("Error in CAnPolygoneReader: subnodePtr->GetnodeType() != MSXML2::NODE_TEXT"); return; }
			}
			if (readM && readText)
			{
				std::vector<double> Container;
				CTextParser<double> TextParser;
				int Count = TextParser.Parse(Container, text);
				if (!TextParser.Good() || Count != 2*M)
				{ enkiIOStruct.Error("Error in CAnPolygoneReader: unable to parse the text"); return; }
				else {
					polygone.clear();
					for (index = 0; index < M; ++index)
						polygone.push_back(An::Point(
							Container[2*index],
							Container[2*index + 1]));
				}
			} else { enkiIOStruct.Error("Error in CAnPolygoneReader: incomplete definition of polygone"); return; }
		} else { enkiIOStruct.Error("Error in CAnPolygoneReader: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CEnkiPhysicalObjectReader class reads Enki::PhysicalObject
	//--------------------------------------------------------------------------------------
	void CEnkiPhysicalObjectReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, Enki::PhysicalObject &physicalObject)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		int index = 0;

		double x = 0.0;
		double y = 0.0;
		double r = 0.0;
		An::Color color;
		double height = 10.0;
		double angle = 0.0;
		double mass = -1.0;
		double staticFrictionThreshold = 0.1;
		double viscousFrictionTau = 0.1;
		double viscousMomentFrictionTau = 0.0;
		double collisionAngularFrictionFactor = 0.01;

		bool readX = false;
		bool readY = false;
		bool readR = false;
		bool readColor = false;
		bool readHeight = false;
		bool readAngle = false;
		bool readMass = false;
		bool readStaticFrictionThreshold = false;
		bool readViscousFrictionTau = false;
		bool readViscousMomentFrictionTau = false;
		bool readCollisionAngularFrictionFactor = false;
		bool readBoundingSurface = false;

		bool readSuccess = true;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "object")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "x") {
						if (!readX) {
							x = atof(attributeValue.c_str());
							if (x > 0.0) readX = true;
							else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: x <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of x"); return; }
					}
					else if (attributeName == "y") {
						if (!readY) {
							y = atof(attributeValue.c_str());
							if (y > 0.0) readY = true;
							else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: y <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of y"); return; }
					}
					else if (attributeName == "r") {
						if (!readR) {
							r = atof(attributeValue.c_str());
							if (r > 0.0) readR = true;
							else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: r <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of r"); return; }
					}
					else if (attributeName == "height") {
						if (!readHeight) {
							height = atof(attributeValue.c_str());
							if (height > 0.0) readHeight = true;
							else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: height <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of height"); return; }
					}
					else if (attributeName == "color") {
						if (!readColor) {
							color = strtocolor(attributeValue);
							readColor = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of color"); return; }
					}
					else if (attributeName == "angle") {
						if (!readAngle) {
							angle = atof(attributeValue.c_str()) * M_PI / 180.0;
							readAngle = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of angle"); return; }
					}
					else if (attributeName == "mass") {
						if (!readMass) {
							mass = atof(attributeValue.c_str());
							readMass = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of mass"); return; }
					}
					else if (attributeName == "staticFrictionThreshold") {
						if (!readStaticFrictionThreshold) {
							staticFrictionThreshold = atof(attributeValue.c_str());
							readStaticFrictionThreshold = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of staticFrictionThreshold"); return; }
					}
					else if (attributeName == "viscousFrictionTau") {
						if (!readViscousFrictionTau) {
							viscousFrictionTau = atof(attributeValue.c_str());
							readViscousFrictionTau = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of viscousFrictionTau"); return; }
					}
					else if (attributeName == "viscousMomentFrictionTau") {
						if (!readViscousMomentFrictionTau) {
							viscousMomentFrictionTau = atof(attributeValue.c_str());
							readViscousMomentFrictionTau = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of viscousMomentFrictionTau"); return; }
					}
					else if (attributeName == "collisionAngularFrictionFactor") {
						if (!readCollisionAngularFrictionFactor) {
							collisionAngularFrictionFactor = atof(attributeValue.c_str());
							readCollisionAngularFrictionFactor = true;
						} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: double definition of collisionAngularFrictionFactor"); return; }
					}
				}
			}
			if (readX && readY)
			{
				physicalObject.pos.x = x;
				physicalObject.pos.y = y;
				physicalObject.r = r;
				physicalObject.color = color;
				physicalObject.height = height;
				physicalObject.angle = angle;
				physicalObject.mass = mass;
				physicalObject.staticFrictionThreshold = staticFrictionThreshold;
				physicalObject.viscousFrictionTau = viscousFrictionTau;
				physicalObject.viscousMomentFrictionTau = viscousMomentFrictionTau;
				physicalObject.collisionAngularFrictionFactor = collisionAngularFrictionFactor;

				physicalObject.angSpeed = 0.0;
				physicalObject.speed = An::Vector(0, 0);

				if (!readR)
				{
					An::Polygone boundingSurface;
					readSuccess = true;
					for (index = 0; index < listPtr->length; ++index)
					{
						subnodePtr = listPtr->Getitem(index);
						if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
						{
							subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
							if (subnodeName == "polygone") {
								if (!readBoundingSurface)
								{
									m_AnPolygoneReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), boundingSurface);
									readSuccess = readBoundingSurface = enkiIOStruct.Good();
								} else readSuccess = false;
							}
							if (!readSuccess)
								break;
						}
						else
							readSuccess = false;
						if (!readSuccess)
							break;
					}
					if (readSuccess && readBoundingSurface)
					{
						physicalObject.setBoundingSurface(boundingSurface);
						physicalObject.initLocalInteractions();
						physicalObject.initGlobalInteractions();
						return;
					}
					else
					{
						enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: unable to read boundingSurface");
						return;
					}
				}
				else
				{
					physicalObject.boundingSurface = false;
					physicalObject.initLocalInteractions();
					physicalObject.initGlobalInteractions();
					return;
				}
			} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: incomplete definition of object"); return; }
		} else { enkiIOStruct.Error("Error in CEnkiPhysicalObjectReader: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CEnkiCircularCamReader class reads Enki::CircularCam
	//--------------------------------------------------------------------------------------
	void CEnkiCircularCamReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, Enki::CircularCam &circularCam)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiCircularCamReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		double x = 0.0;
		double y = 0.0;
		double height = 0.0;
		double orientation = 0.0;
		double fieldOfView = 0.0;
		int    pixelCount = 0;

		bool readX = false;
		bool readY = false;
		bool readHeight = false;
		bool readOrientation = false;
		bool readFieldOfView = false;
		bool readPixelCount = false;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "camera")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "x") {
						if (!readX) {
							x = atof(attributeValue.c_str());
							readX = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of x"); return; }
					}
					else if (attributeName == "y") {
						if (!readY) {
							y = atof(attributeValue.c_str());
							readY = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of y"); return; }
					}
					else if (attributeName == "height") {
						if (!readHeight) {
							height = atof(attributeValue.c_str());
							if (height > 0.0) readHeight = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: height <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of height"); return; }
					}
					else if (attributeName == "orientation") {
						if (!readOrientation) {
							orientation = atof(attributeValue.c_str()) * M_PI / 180.0;
							readOrientation = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of orientation"); return; }
					}
					else if (attributeName == "fieldOfView") {
						if (!readFieldOfView) {
							fieldOfView = atof(attributeValue.c_str()) * M_PI / 180.0;
							if (fieldOfView > 0.0) readFieldOfView = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: fieldOfView <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of fieldOfView"); return; }
					}
					else if (attributeName == "pixelCount") {
						if (!readPixelCount) {
							pixelCount = atoi(attributeValue.c_str());
							if (pixelCount > 0) readPixelCount = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: pixelCount <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of pixelCount"); return; }
					}
				}
			}
			if (readX && readY && readHeight && readOrientation && readFieldOfView && readPixelCount)
			{
				circularCam.setCircularCam(An::Vector(x, y), height, orientation, fieldOfView, pixelCount);
				return;
			} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: incomplete definition of camera"); return; }
		} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CEnkiIRSensorReader class reads Enki::IRSensor
	//--------------------------------------------------------------------------------------
	void CEnkiIRSensorReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, Enki::IRSensor &irSensor)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiIRSensorReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		double x = 0.0;
		double y = 0.0;
		double height = 0.0;
		double orientation = 0.0;
		double range = 0.0;
		double aperture = 0.0;
		/*int    rayCount = 0;
		std::string sensorResponseKernel;*/

		bool readX = false;
		bool readY = false;
		bool readHeight = false;
		bool readOrientation = false;
		bool readRange = false;
		bool readAperture = false;
		/*bool readRayCount = false;
		bool readSensorResponseKernel = false;*/

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "irsensor")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "x") {
						if (!readX) {
							x = atof(attributeValue.c_str());
							readX = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of x"); return; }
					}
					else if (attributeName == "y") {
						if (!readY) {
							y = atof(attributeValue.c_str());
							readY = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of y"); return; }
					}
					else if (attributeName == "height") {
						if (!readHeight) {
							height = atof(attributeValue.c_str());
							if (height > 0.0) readHeight = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: height <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of height"); return; }
					}
					else if (attributeName == "orientation") {
						if (!readOrientation) {
							orientation = atof(attributeValue.c_str()) * M_PI / 180.0;
							readOrientation = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of orientation"); return; }
					}
					else if (attributeName == "range") {
						if (!readRange) {
							range = atof(attributeValue.c_str());
							readRange = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of range"); return; }
					}
					else if (attributeName == "aperture") {
						if (!readAperture) {
							aperture = atof(attributeValue.c_str());
							readAperture = true;
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of aperture"); return; }
					}
					/*else if (attributeName == "rayCount") {
						if (!readRayCount) {
							rayCount = atoi(attributeValue.c_str());
							if (rayCount == 1) readRayCount = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: rayCount <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of rayCount"); return; }
					}
					else if (attributeName == "sensorResponseKernel") {
						if (!readSensorResponseKernel) {
							sensorResponseKernel = attributeValue.c_str();
							if (sensorResponseKernel == "simple")
								readSensorResponseKernel = true;
							else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: unknown sensorResponseKernel"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: double definition of sensorResponseKernel"); return; }
					}*/
				}
			}
			if (readX && readY && readHeight && readOrientation && readRange)
			{
				irSensor.setIRSensor(An::Vector(x, y), height, orientation, range, aperture);
				return;
			} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: incomplete definition of irsensor"); return; }
		} else { enkiIOStruct.Error("Error in CEnkiCircularCamReader: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CEnkiWorldIO class implements IEnkiIO interface for CEnkiWorld
	//--------------------------------------------------------------------------------------
	void CEnkiWorldIO::Read(RFW::CEnkiIOStruct &enkiIOStruct)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiWorldIO: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		int index = 0;

		double width        = 0.0;
		double height       = 0.0;
		double startX       = 0.0;
		double startY       = 0.0;
		double startAngle   = 0.0;

		bool readWidth      = false;
		bool readHeight     = false;
		bool readStart      = false;
		bool readStartX     = false;
		bool readStartY     = false;
		bool readStartAngle = false;
		bool readSuccess    = true;

		CEnkiPhysicalObject * physicalObject = 0;

		m_EnkiWorld.ClearWorld();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "world")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "width") {
						if (!readWidth) {
							width = atof(attributeValue.c_str());
							if (width > 0.0) readWidth = true;
							else { enkiIOStruct.Error("Error in CEnkiWorldIO: width <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiWorldIO: double definition of width"); return; }
					}
					else if (attributeName == "height") {
						if (!readHeight) {
							height = atof(attributeValue.c_str());
							if (height > 0.0) readHeight = true;
							else { enkiIOStruct.Error("Error in CEnkiWorldIO: height <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiWorldIO: double definition of height"); return; }
					}
					else if (attributeName == "startX") {
						if (!readStartX) {
							startX = atof(attributeValue.c_str());
							if (startX > 0.0) readStartX = true;
							else { enkiIOStruct.Error("Error in CEnkiWorldIO: startX <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiWorldIO: double definition of startX"); return; }
					}
					else if (attributeName == "startY") {
						if (!readStartY) {
							startY = atof(attributeValue.c_str());
							if (startY > 0.0) readStartY = true;
							else { enkiIOStruct.Error("Error in CEnkiWorldIO: startY <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiWorldIO: double definition of startY"); return; }
					}
					else if (attributeName == "startAngle") {
						if (!readStartAngle) {
							startAngle = atof(attributeValue.c_str()) * M_PI / 180.0;
							readStartAngle = true;
						} else { enkiIOStruct.Error("Error in CEnkiWorldIO: double definition of startAngle"); return; }
					}
				}
			}
			if (readWidth && readHeight)
			{
				m_EnkiWorld.ClearWorld(width, height);

				if (readStartX && readStartY &&
					startX > 0.0 && startX < width &&
					startY > 0.0 && startY < height)
				{
					/*if (readStartAngle)
						m_StartAngle = startAngle;
					else m_StartAngle = 0.0;*/
					readStart = true;

					m_EnkiWorld.AddStartingPoint(An::Point(startX, startY));
				}
				else readStart = false;

				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
					{
						subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
						if (subnodeName == "object") {
							physicalObject = new CEnkiPhysicalObject(m_EnkiWorld);
							physicalObject->Read(enkiIOStruct.SetNodePtr(subnodePtr));
							if (!enkiIOStruct.Good()) {
								enkiIOStruct.Error("Error in CEnkiWorldIO: unable to read object");
								readSuccess = false;
								break;
							}
						}
					}
					else
					{
						enkiIOStruct.Error("Error in CEnkiWorldIO: subnodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
						readSuccess = false;
						break;
					}
				}

				if (!readSuccess)
				{
					m_EnkiWorld.ClearWorld();
					return;
				}
			} else { enkiIOStruct.Error("Error in CEnkiWorldIO: incomplete definition of world"); return; }
		} else { enkiIOStruct.Error("Error in CEnkiWorldIO: bad nodeName"); return; }
	}

	//--------------------------------------------------------------------------------------
	// CEnkiPhysicalObjectIO class implements IEnkiIO interface for CEnkiPhysicalObject
	//--------------------------------------------------------------------------------------
	void CEnkiPhysicalObjectIO::Read(RFW::CEnkiIOStruct &enkiIOStruct)
	{
		m_EnkiPhysicalObject.ClearPhysicalObject();
		m_EnkiPhysicalObjectReader.Read(enkiIOStruct, *m_EnkiPhysicalObject.GetEnkiPhysicalObject());
		if (!enkiIOStruct.Good())
		{
			enkiIOStruct.Error("Error in CEnkiPhysicalObjectIO: unable to read object");
			return;
		}
	}

	//--------------------------------------------------------------------------------------
	// CEnkiRobotIO class implements IEnkiIO interface for CEnkiRobot
	//--------------------------------------------------------------------------------------
	void CEnkiRobotIO::Read(RFW::CEnkiIOStruct &enkiIOStruct)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiRobotIO: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		int index = 0;

		bool readObject  = false;
		bool readSuccess = true;

		m_EnkiRobot.ClearRobot();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "robot")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());

					// TODO: Here you should process the atributes of the robot

				}
			}
			/*if ( CONDITION )*/
			{
				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
					{
						subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
						if (subnodeName == "object") {
							if (!readObject) {
								m_EnkiPhysicalObjectReader.Read(
									enkiIOStruct.SetNodePtr(subnodePtr), *m_EnkiRobot.GetEnkiRobot());
								if (!enkiIOStruct.Good()) {
									enkiIOStruct.Error("Error in CEnkiRobotIO: unable to read object");
									readSuccess = readObject = false;
									break;
								}
								else readSuccess = readObject = true;
							} else {
								enkiIOStruct.Error("Error in CEnkiRobotIO: double definition of object");
								readSuccess = readObject = false;
								break;
							}
						}

						// TODO: Here you should add additional conditions

					}
					else readSuccess = false;
					if (!readSuccess)
						break;
				}
				if (!readSuccess || !readObject)
				{
					m_EnkiRobot.ClearRobot();
					return;
				}
			} /* else { enkiIOStruct.Error("Error in CEnkiRobotIO: incomplete definition of robot"); return; } */
		} else { enkiIOStruct.Error("Error in CEnkiRobotIO: bad nodeName"); return; }
	}
}