//--------------------------------------------------------------------------------------
// File: RFWNeuralNetwork.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core classes extension source
// Implementation of CFeedForward class
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)
#pragma warning(disable : 4786)

#include <cassert>
#include <cmath>
#include <string>
#include <vector>

#include "RFWAlgebra.h"

#include "RFWNeuralNetwork.h"
#include "RFWNeuralNetworkIO.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CFeedForward class for simulating feed-forward neural networks
	//--------------------------------------------------------------------------------------
	double CFeedForward::logsig(double x)
	{
		return 1.0 / (1.0 + exp(-x));
	}

	double CFeedForward::tansig(double x)
	{
		return 2.0 / (1.0 + exp(-2.0 * x)) - 1.0;
	}

	double CFeedForward::purelin(double x)
	{
		return x;
	}

	bool CFeedForward::CheckConsistency() const
	{
		bool checkMean = false;
		bool checkStd  = false;
		bool checkPCAMatrix = false;
		bool checkWeights = false;
		bool checkBiases  = false;
		bool checkLayers  = false;
		std::vector<int> dimensions;
		int i = 0;

		if (m_readMean)
			checkMean = (m_Mean.GetM() == m_InputDimensions);
		else checkMean = true;

		if (m_readStd)
			checkStd = (m_Std.GetM() == m_InputDimensions);
		else checkStd = true;

		if (m_readPCAMatrix)
		{
			dimensions.push_back(m_PCAMatrix.GetM());
			checkPCAMatrix = (m_PCAMatrix.GetN() == m_InputDimensions);
		}
		else
		{
			dimensions.push_back(m_InputDimensions);
			checkPCAMatrix = true;
		}

		if (!checkMean || !checkStd || !checkPCAMatrix)
			return false;

		for (i = 0; i < m_NumLayers; ++i)
		{
			dimensions.push_back(m_Weights[i].GetN());
			dimensions.push_back(m_Weights[i].GetM());
		}

		dimensions.push_back(m_OutputDimensions);

		checkWeights = true;
		for (i = 0; i <= m_NumLayers; ++i)
			if (dimensions[2*i] != dimensions[2*i + 1])
			{
				checkWeights = false;
				break;
			}

		checkBiases = true;
		for (i = 0; i < m_NumLayers; ++i)
			if (m_Biases[i].GetM() != m_Weights[i].GetM())
			{
				checkBiases = false;
				break;
			}

		checkLayers = true;
		for (i = 0; i < m_NumLayers; ++i)
			if (m_LayersDimensions[i] != m_Weights[i].GetM())
			{
				checkLayers = false;
				break;
			}

		return checkWeights && checkBiases && checkLayers;
	}

	CVector CFeedForward::Simulate(const CVector & x) const
	{
		CVector xmean;
		CVector xstd;
		CVector y;
		int Layer = 0;

		if (x.GetM() == m_InputDimensions)
		{
			if (m_readMean)
				xmean = x - m_Mean;
			else xmean = x;
			if (m_readStd)
				xstd  = xmean / m_Std;
			else xstd = xmean;
			if (m_readPCAMatrix)
				y = m_PCAMatrix * xstd;
			else y = xstd;

			for (Layer = 0; Layer < m_NumLayers; ++Layer)
			{
				y = m_Weights[Layer] * y + m_Biases[Layer];
				y.Apply(m_LayersTransferFcn[Layer]);
			}
		}

		return y;
	}

	void CFeedForward::Clear()
	{
		int i = 0;
		m_Mean.Clear();
		m_Std.Clear();
		m_PCAMatrix.Clear();
		m_LayersDimensions.clear();
		m_LayersTransferFcn.clear();
		for (i = 0; i < m_NumLayers; ++i)
		{
			m_Weights[i].Clear();
			m_Biases[i].Clear();
		}
		m_Weights.clear();
		m_Biases.clear();
		m_InputDimensions = 0;
		m_OutputDimensions = 0;
		m_NumLayers = 0;
		m_readMean      = false;
		m_readStd       = false;
		m_readPCAMatrix = false;
	}
}