//--------------------------------------------------------------------------------------
// File: RFWNeuralNetworkIO.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core IO extension source
// Implementation of CFeedForwardIO class
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)

#include "TextParser.h"

#include "RFWAlgebra.h"
#include "RFWAlgebraIO.h"

#include "RFWNeuralNetwork.h"
#include "RFWNeuralNetworkIO.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CFeedForwardReader class reads RFW::CFeedForward
	//--------------------------------------------------------------------------------------
	CFeedForward::function * CFeedForwardReader::GetTransferFcn(std::string TransferFcn)
	{
		if (TransferFcn == "logsig")
			return &CFeedForward::logsig;
		else if (TransferFcn == "tansig")
			return &CFeedForward::tansig;
		else if (TransferFcn == "purelin")
			return &CFeedForward::purelin;
		else
		{
			assert(false);
			return 0;
		}
	}

	bool CFeedForwardReader::CheckTransferFcn(std::string TransferFcn)
	{
		return
			(TransferFcn == "logsig" ||
			 TransferFcn == "tansig" ||
			 TransferFcn == "purelin");
	}

	void CFeedForwardReader::Read(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		int index = 0;

		int inputDimensions = 0;
		int outputDimensions = 0;
		int numLayers = 0;
		bool readInputDimensions = false;
		bool readOutputDimensions = false;
		bool readNumLayers = false;
		bool readLayers    = false;
		bool readWeights   = false;
		bool readBiases    = false;
		bool readSuccess   = true;

		feedForward.Clear();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "network")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "inputDimensions") {
						if (!readInputDimensions) {
							inputDimensions = atoi(attributeValue.c_str());
							if (inputDimensions > 0) readInputDimensions = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: inputDimensions <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of inputDimensions"); return; }
					}
					else if (attributeName == "outputDimensions") {
						if (!readOutputDimensions) {
							outputDimensions = atoi(attributeValue.c_str());
							if (outputDimensions > 0) readOutputDimensions = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: outputDimensions <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of outputDimensions"); return; }
					}
					else if (attributeName == "numLayers") {
						if (!readNumLayers) {
							numLayers = atoi(attributeValue.c_str());
							if (numLayers > 0) readNumLayers = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: numLayers <= 0"); return; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of numLayers"); return; }
					}
				}
			}
			if (readInputDimensions && readOutputDimensions && readNumLayers)
			{
				feedForward.m_InputDimensions = inputDimensions;
				feedForward.m_OutputDimensions = outputDimensions;
				feedForward.m_NumLayers = numLayers;

				feedForward.m_LayersDimensions.resize(numLayers);
				feedForward.m_LayersTransferFcn.resize(numLayers);
				feedForward.m_Weights.resize(numLayers);
				feedForward.m_Biases.resize(numLayers);

				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
					{
						subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
						if (subnodeName == "mean")
						{
							if (!feedForward.m_readMean)
								readSuccess = feedForward.m_readMean =
								  ReadMean(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						else if (subnodeName == "std")
						{
							if (!feedForward.m_readStd)
								readSuccess = feedForward.m_readStd =
								  ReadStd(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						else if (subnodeName == "PCAmatrix")
						{
							if (!feedForward.m_readPCAMatrix)
								readSuccess = feedForward.m_readPCAMatrix =
								  ReadPCAMatrix(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						else if (subnodeName == "layers")
						{
							if (!readLayers)
								readSuccess = readLayers =
								  ReadLayers(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						else if (subnodeName == "weights")
						{
							if (!readWeights)
								readSuccess = readWeights =
								  ReadWeights(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						else if (subnodeName == "biases")
						{
							if (!readBiases)
								readSuccess = readBiases =
								  ReadBiases(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
							else readSuccess = false;
						}
						if (!readSuccess)
							break;
					}
					else
					{
						enkiIOStruct.Error("Error in CFeedForwardReader: subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT");
						readSuccess = false;
						break;
					}
				}
				bool networkConsistency = feedForward.CheckConsistency();
				if (!readSuccess || !readLayers ||
					!readWeights || !readBiases ||
					!networkConsistency)
				{
					feedForward.Clear();
					enkiIOStruct.Error("Error in CFeedForwardReader: unable to read network");
					if (!networkConsistency)
						enkiIOStruct.Error("Error in CFeedForwardReader: inconsistent network");
					return;
				}
				else return;
			} else  { enkiIOStruct.Error("Error in CFeedForwardReader: incomplete definition of network"); return; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName"); return; }
	}

	bool CFeedForwardReader::ReadMean(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "mean")
		{
			if (listPtr->length == 1)
			{
				subnodePtr = listPtr->Getitem(0);
				m_VectorReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), feedForward.m_Mean);
				if (!enkiIOStruct.Good()) {
					enkiIOStruct.Error("Error in CFeedForwardReader: unable to read mean");
					return false;
				} else return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of mean"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of mean"); return false; }
	}

	bool CFeedForwardReader::ReadStd(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "std")
		{
			if (listPtr->length == 1)
			{
				subnodePtr = listPtr->Getitem(0);
				m_VectorReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), feedForward.m_Std);
				if (!enkiIOStruct.Good()) {
					enkiIOStruct.Error("Error in CFeedForwardReader: unable to read std");
					return false;
				} else return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of std"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of std"); return false; }
	}

	bool CFeedForwardReader::ReadPCAMatrix(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "PCAmatrix")
		{
			if (listPtr->length == 1)
			{
				subnodePtr = listPtr->Getitem(0);
				m_MatrixReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), feedForward.m_PCAMatrix);
				if (!enkiIOStruct.Good()) {
					enkiIOStruct.Error("Error in CFeedForwardReader: unable to read PCAmatrix");
					return false;
				} else return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of PCAmatrix"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of PCAmatrix"); return false; }
	}

	bool CFeedForwardReader::ReadLayers(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		int index = 0;

		int layerIndex = 0;
		std::vector<bool> readLayer;
		readLayer.resize(feedForward.m_NumLayers, false);

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "layers")
		{
			if (feedForward.m_NumLayers == listPtr->length)
			{
				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					layerIndex = ReadLayer(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
					if (!enkiIOStruct.Good() || layerIndex == 0)
					  { enkiIOStruct.Error("Error in CFeedForwardReader: unable to read layers"); return false; }
					else readLayer[layerIndex - 1] = true;
				}
				for (index = 0; index < feedForward.m_NumLayers; ++index)
					if (readLayer[index] == false)
						{ enkiIOStruct.Error("Error in CFeedForwardReader: inconsistent layers"); return false; }
				return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of layers"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of layers"); return false; }
	}

	int CFeedForwardReader::ReadLayer(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return 0;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;

		int layerIndex = 0;
		int layerDimensions = 0;
		CFeedForward::function * layerTransferFcn = 0;
		bool readIndex = false;
		bool readDimensions = false;
		bool readTransferFcn = false;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "layer")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "index") {
						if (!readIndex) {
							layerIndex = atoi(attributeValue.c_str());
							if (layerIndex > 0 && layerIndex <= feedForward.m_NumLayers) readIndex = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: bad layer index"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of layer index"); return 0; }
					}
					else if (attributeName == "dimensions") {
						if (!readDimensions) {
							layerDimensions = atoi(attributeValue.c_str());
							if (layerDimensions > 0) readDimensions = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: dimensions <= 0"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of dimensions"); return 0; }
					}
					else if (attributeName == "transferFcn") {
						if (!readTransferFcn) {
							if (CheckTransferFcn(attributeValue))
							{
								layerTransferFcn = GetTransferFcn(attributeValue);
								readTransferFcn = true;
							} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad transferFcn"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of transferFcn"); return 0; }
					}
				}
			}
			if (readIndex && readDimensions && readTransferFcn)
			{
				feedForward.m_LayersDimensions[layerIndex - 1] = layerDimensions;
				feedForward.m_LayersTransferFcn[layerIndex - 1] = layerTransferFcn;
				return layerIndex;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: incomplete definition of layer"); return 0; }
		} else { enkiIOStruct.Error("Error in CMatrixReader: bad nodeName of layer"); return 0; }
	}

	bool CFeedForwardReader::ReadWeights(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		int index = 0;

		int weightMatrixIndex = 0;
		std::vector<bool> readWeightMatrix;
		readWeightMatrix.resize(feedForward.m_NumLayers, false);

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "weights")
		{
			if (feedForward.m_NumLayers == listPtr->length)
			{
				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					weightMatrixIndex = ReadWeight(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
					if (!enkiIOStruct.Good() || weightMatrixIndex == 0)
						{ enkiIOStruct.Error("Error in CFeedForwardReader: unable to read weights"); return false; }
					else readWeightMatrix[weightMatrixIndex - 1] = true;
				}
				for (index = 0; index < feedForward.m_NumLayers; ++index)
					if (readWeightMatrix[index] == false)
						{ enkiIOStruct.Error("Error in CFeedForwardReader: inconsistent weights"); return false; }
				return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of weights"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of weights"); return false; }
	}

	int CFeedForwardReader::ReadWeight(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return 0;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;

		int toLayer = 0;
		int fromLayer = 0;
		bool readTo = false;
		bool readFrom = false;
		bool readMatrix = false;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "inputlayer" || nodeName == "hiddenlayer")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "to") {
						if (!readTo) {
							toLayer = atoi(attributeValue.c_str());
							if (toLayer > 0 && toLayer <= feedForward.m_NumLayers) readTo = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: bad to layer index"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of to layer index"); return 0; }
					}
					else if (attributeName == "from") {
						if (!readFrom) {
							fromLayer = atoi(attributeValue.c_str());
							if (fromLayer >= 0 && fromLayer < feedForward.m_NumLayers) readFrom = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: bad from layer index"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of from layer index"); return 0; }
					}
				}
			}
			if (readTo && toLayer == fromLayer + 1)
			{
				if (listPtr->length == 1)
				{
					subnodePtr = listPtr->Getitem(0);
					m_MatrixReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), feedForward.m_Weights[toLayer - 1]);
					if (enkiIOStruct.Good())
						readMatrix = true;
					else { enkiIOStruct.Error("Error in CFeedForwardReader: unable to read weight matrix"); return 0; }
				} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of weight matrix"); return 0; }
				if (readMatrix)
					return toLayer;
				else return 0;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: incomplete definition of weight matrix"); return 0; }
		} else { enkiIOStruct.Error("Error in CMatrixReader: bad nodeName of weight matrix"); return 0; }
	}

	bool CFeedForwardReader::ReadBiases(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return false;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;

		int index = 0;

		int biasIndex = 0;
		std::vector<bool> readBias;
		readBias.resize(feedForward.m_NumLayers, false);

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "biases")
		{
			if (feedForward.m_NumLayers == listPtr->length)
			{
				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					biasIndex = ReadBias(enkiIOStruct.SetNodePtr(subnodePtr), feedForward);
					if (!enkiIOStruct.Good() || biasIndex == 0)
						{ enkiIOStruct.Error("Error in CFeedForwardReader: unable to read biases"); return false; }
					else readBias[biasIndex - 1] = true;
				}
				for (index = 0; index < feedForward.m_NumLayers; ++index)
					if (readBias[index] == false)
						{ enkiIOStruct.Error("Error in CFeedForwardReader: inconsistent biases"); return false; }
				return true;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of biases"); return false; }
		} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad nodeName of biases"); return false; }
	}

	int CFeedForwardReader::ReadBias(RFW::CEnkiIOStruct &enkiIOStruct, RFW::CFeedForward &feedForward)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CFeedForwardReader: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return 0;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string attributeName;
		std::string attributeValue;

		int toLayer = 0;
		bool readTo = false;
		bool readVector = false;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "bias")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "to") {
						if (!readTo) {
							toLayer = atoi(attributeValue.c_str());
							if (toLayer > 0 && toLayer <= feedForward.m_NumLayers) readTo = true;
							else { enkiIOStruct.Error("Error in CFeedForwardReader: bad to layer index"); return 0; }
						} else { enkiIOStruct.Error("Error in CFeedForwardReader: double definition of to layer index"); return 0; }
					}
				}
			}
			if (readTo)
			{
				if (listPtr->length == 1)
				{
					subnodePtr = listPtr->Getitem(0);
					m_VectorReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), feedForward.m_Biases[toLayer - 1]);
					if (enkiIOStruct.Good())
						readVector = true;
					else { enkiIOStruct.Error("Error in CFeedForwardReader: unable to read bias vector"); return 0; }
				} else { enkiIOStruct.Error("Error in CFeedForwardReader: bad definition of bias vector"); return 0; }
				if (readVector)
					return toLayer;
				else return 0;
			} else { enkiIOStruct.Error("Error in CFeedForwardReader: incomplete definition of bias vector"); return 0; }
		} else { enkiIOStruct.Error("Error in CMatrixReader: bad nodeName of bias vector"); return 0; }
	}
}