	//--------------------------------------------------------------------------------------
// File: RFWNeuralRobotGraphics.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core graphics extension source
// Implementation of CEnkiNeuralRobotGraphics class
//
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning(disable : 4530)

#include "enki/PhysicalEngine.h"
#include "enki/Interaction.h"
#include "enki/interactions/CircularCam.h"
#include "enki/interactions/IRSensor.h"
#include "enki/robots/NeuralRobot.h"

#include "CustomVertex.h"
#include "VertexBuffer.h"

#include "RFWNeuralRobot.h"
#include "RFWNeuralRobotGraphics.h"

namespace RFW
{
	//---------------------------------------------------------------------------------------
	// CEnkiNeuralRobotGraphics class implements IEnkiGraphics interface for CEnkiNeuralRobot
	//---------------------------------------------------------------------------------------
	CEnkiNeuralRobotGraphics::CEnkiNeuralRobotGraphics(RFW::CEnkiNeuralRobot &enkiNeuralRobot):
		m_VertexBuffer(enkiNeuralRobot.m_VertexBuffer),
		m_EnkiNeuralRobot(enkiNeuralRobot),
		m_EnkiPhysicalObjectPlotter(enkiNeuralRobot.m_VertexBuffer),
		m_EnkiCircularCamPlotter(enkiNeuralRobot.m_VertexBuffer),
		m_EnkiIRSensorPlotter(enkiNeuralRobot.m_VertexBuffer)
	{
	
	}
	
	void CEnkiNeuralRobotGraphics::Render()
	{
		m_EnkiPhysicalObjectPlotter.Render(*m_EnkiNeuralRobot.GetEnkiRobot());
		RenderDirection();
		RenderWheels();

		size_t CamerasCount = m_EnkiNeuralRobot.GetEnkiRobot()->cameras.size();
		size_t IRSensorsCount = m_EnkiNeuralRobot.GetEnkiRobot()->infraredSensors.size();
		size_t index = 0;

		for (index = 0; index < CamerasCount; ++index)
			m_EnkiCircularCamPlotter.Render(*m_EnkiNeuralRobot.GetEnkiRobot()->cameras[index],
				m_EnkiNeuralRobot.GetEnkiRobot()->r);

		for (index = 0; index < IRSensorsCount; ++index)
			m_EnkiIRSensorPlotter.Render(*m_EnkiNeuralRobot.GetEnkiRobot()->infraredSensors[index]);
	}

	void CEnkiNeuralRobotGraphics::RenderDirection()
	{
		float width  = (float)m_EnkiNeuralRobot.GetEnkiRobot()->r / 2.0f;
		float height = (float)m_EnkiNeuralRobot.GetEnkiRobot()->r / 2.0f;
		float angle  = (float)m_EnkiNeuralRobot.GetEnkiRobot()->angle;

		CUSTOMVERTEX vertices[3] = 
		{
			{-height, -width, 0.4f, 0xffaaaaaa, }, // x, y, z, color
			{-height, +width, 0.4f, 0xffaaaaaa, },
			{+height,   0.0f, 0.4f, 0xffaaaaaa, },
		};

		m_VertexBuffer.CopyVertices(vertices, 3 * sizeof(CUSTOMVERTEX));
		m_VertexBuffer.SetVertexBuffer();

		D3DXMATRIX matTranslate;
		D3DXMATRIX matRotate;
		D3DXMATRIX matTransform;

		D3DXMatrixTranslation(&matTranslate,
			(float)m_EnkiNeuralRobot.GetEnkiRobot()->pos.x,
			(float)m_EnkiNeuralRobot.GetEnkiRobot()->pos.y, 0.0f);
		D3DXMatrixRotationZ  (&matRotate, angle);
		D3DXMatrixMultiply(&matTransform, &matRotate, &matTranslate);

		m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLELIST, matTransform);
	}

	void CEnkiNeuralRobotGraphics::RenderWheels()
	{
		float length   = (float)m_EnkiNeuralRobot.GetEnkiRobot()->r;
		float width    = (float)m_EnkiNeuralRobot.GetEnkiRobot()->r / 4.0f;
		float distance = (float)m_EnkiNeuralRobot.GetEnkiRobot()->wheelDist / 2.0f;
		float angle    = (float)m_EnkiNeuralRobot.GetEnkiRobot()->angle;

		CUSTOMVERTEX vertices[4] = 
		{
			{   -length / 2.0f, +width / 2.0f, 0.6f, 0xffaaaaaa, }, // x, y, z, color
			{   -length / 2.0f, -width / 2.0f, 0.6f, 0xffaaaaaa, },
			{   +length / 2.0f, +width / 2.0f, 0.6f, 0xffaaaaaa, },
			{   +length / 2.0f, -width / 2.0f, 0.6f, 0xffaaaaaa, },
		};

		m_VertexBuffer.CopyVertices(vertices, 4 * sizeof(CUSTOMVERTEX));
		m_VertexBuffer.SetVertexBuffer();

		D3DXMATRIX matWorldLeft;
		D3DXMATRIX matWorldRight;
		D3DXMATRIX matRotate;
		D3DXMATRIX matTranslateLeft;
		D3DXMATRIX matTranslateRight;
		D3DXMATRIX matTranslate;

		D3DXMatrixTranslation(&matTranslateLeft , 0.0f, +distance, 0.0f);
		D3DXMatrixTranslation(&matTranslateRight, 0.0f, -distance, 0.0f);
		D3DXMatrixTranslation(&matTranslate,
			(float)m_EnkiNeuralRobot.GetEnkiRobot()->pos.x,
			(float)m_EnkiNeuralRobot.GetEnkiRobot()->pos.y, 0.0f);
		D3DXMatrixRotationZ  (&matRotate, angle);
		D3DXMatrixMultiply(&matWorldLeft , &matTranslateLeft , &matRotate);
		D3DXMatrixMultiply(&matWorldRight, &matTranslateRight, &matRotate);
		D3DXMatrixMultiply(&matWorldLeft , &matWorldLeft     , &matTranslate);
		D3DXMatrixMultiply(&matWorldRight, &matWorldRight    , &matTranslate);

		m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLESTRIP, matWorldLeft);
		m_VertexBuffer.DrawPrimitive(D3DPT_TRIANGLESTRIP, matWorldRight);
	}
}