//--------------------------------------------------------------------------------------
// File: RFWCoreIO.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Framework core IO extension source
// Implementation of CEnkiNeuralRobotIO class
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)

#include "enki/PhysicalEngine.h"
#include "enki/Interaction.h"
#include "enki/interactions/CircularCam.h"
#include "enki/interactions/IRSensor.h"
#include "enki/robots/NeuralRobot.h"

#include "TextParser.h"

#include "RFWNeuralRobot.h"
#include "RFWNeuralRobotIO.h"

namespace RFW
{
	//--------------------------------------------------------------------------------------
	// CEnkiNeuralRobotIO class implements IEnkiIO interface for CEnkiNeuralRobot
	//--------------------------------------------------------------------------------------
	void CEnkiNeuralRobotIO::Read(RFW::CEnkiIOStruct &enkiIOStruct)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr         subnodePtr;
		MSXML2::IXMLDOMNamedNodeMapPtr attributeMapPtr;
		MSXML2::IXMLDOMNodePtr         attributePtr;
		MSXML2::IXMLDOMNodeListPtr     listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string attributeName;
		std::string attributeValue;

		int index = 0;

		double wheelDistance = 0.0;
		double robotSpeed = 0.0;

		bool readWheelDistance = false;
		bool readRobotSpeed = false;
		bool readObject  = false;
		bool readSensors = false;
		bool readSuccess = true;

		m_EnkiNeuralRobot.ClearRobot();

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "neuralrobot")
		{
			attributeMapPtr = nodePtr->Getattributes();
			attributeMapPtr->reset();
			while (attributePtr = attributeMapPtr->nextNode())
			{
				if (attributePtr->GetnodeType() == MSXML2::NODE_ATTRIBUTE)
				{
					attributeName  = (const char *)bstr_t(attributePtr->GetnodeName());
					attributeValue = (const char *)bstr_t(attributePtr->GetnodeValue());
					if (attributeName == "wheelDistance") {
						if (!readWheelDistance) {
							wheelDistance = atof(attributeValue.c_str());
							if (wheelDistance > 0.0) readWheelDistance = true;
							else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: wheelDistance <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: double definition of wheelDistance"); return; }
					}
					else if (attributeName == "robotSpeed") {
						if (!readRobotSpeed) {
							robotSpeed = atof(attributeValue.c_str());
							if (robotSpeed > 0.0) readRobotSpeed = true;
							else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: robotSpeed <= 0.0"); return; }
						} else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: robotSpeed <= 0.0"); return; }
					}
				}
			}
			if (readWheelDistance && readRobotSpeed)
			{
				m_EnkiNeuralRobot.GetEnkiRobot()->wheelDist = wheelDistance;
				m_EnkiNeuralRobot.GetEnkiRobot()->robotSpeed = robotSpeed;

				for (index = 0; index < listPtr->length; ++index)
				{
					subnodePtr = listPtr->Getitem(index);
					if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
					{
						subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
						if (subnodeName == "object") {
							if (!readObject) {
								m_EnkiPhysicalObjectReader.Read(
									enkiIOStruct.SetNodePtr(subnodePtr), *m_EnkiNeuralRobot.GetEnkiRobot());
								if (!enkiIOStruct.Good()) {
									enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: unable to read object");
									readSuccess = readObject = false;
									break;
								}
								else readSuccess = readObject = true;
							} else {
								enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: double definition of object");
								readSuccess = readObject = false;
								break;
							}
						}
						else if (subnodeName == "sensors") {
							if (!readSensors) {
								ReadSensors(enkiIOStruct.SetNodePtr(subnodePtr));
								if (!enkiIOStruct.Good()) {
									enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: unable to read sensors");
									readSuccess = readSensors = false;
									break;
								}
								else readSuccess = readSensors = true;
							} else {
								enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: double definition of sensors");
								readSuccess = readSensors = false;
								break;
							}
						}
					}
					else readSuccess = false;
					if (!readSuccess)
						break;
				}
				if (readSuccess && readObject)
				{
					m_EnkiNeuralRobot.GetEnkiRobot()->initLocalInteractions();
					m_EnkiNeuralRobot.GetEnkiRobot()->initGlobalInteractions();
					return;
				}
				else
				{
					m_EnkiNeuralRobot.ClearRobot();
					return;
				}
			} else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: incomplete definition of neuralrobot"); return; }
		} else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: bad nodeName"); return; }
	}

	void CEnkiNeuralRobotIO::ReadSensors(RFW::CEnkiIOStruct &enkiIOStruct)
	{
		MSXML2::IXMLDOMNodePtr nodePtr = enkiIOStruct.GetNodePtr();

		if (nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT) {
			enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: nodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
			return;
		}

		MSXML2::IXMLDOMNodePtr     subnodePtr;
		MSXML2::IXMLDOMNodeListPtr listPtr = nodePtr->childNodes;

		std::string nodeName;
		std::string subnodeName;
		std::string text;

		int index = 0;

		Enki::IRSensor    * irSensor = 0;
		Enki::CircularCam * Camera   = 0;

		bool readSuccess = true;

		nodeName = (const char *)bstr_t(nodePtr->GetnodeName());
		if (nodeName == "sensors")
		{
			for (index = 0; index < listPtr->length; ++index)
			{
				subnodePtr = listPtr->Getitem(index);
				if (subnodePtr->GetnodeType() == MSXML2::NODE_ELEMENT)
				{
					subnodeName = (const char *)bstr_t(subnodePtr->GetnodeName());
					if (subnodeName == "irsensor") {
						// WARNING: Implementation dependent code
						irSensor = new Enki::IRSensor(1, &Enki::neuralRobotIRSensorModelPtr); 
						m_EnkiIRSensorReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), *irSensor);
						if (enkiIOStruct.Good())
						{
							irSensor->setOwner(m_EnkiNeuralRobot.GetEnkiRobot());
							m_EnkiNeuralRobot.GetEnkiRobot()->infraredSensors.push_back(irSensor);
							m_EnkiNeuralRobot.GetEnkiRobot()->addLocalInteraction(irSensor);
							readSuccess = true;
						}
						else
						{
							enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: unable to read irsensor");
							readSuccess = false;
							delete irSensor;
							break;
						}
					}
					else if (subnodeName == "camera") {
						// WARNING: Implementation dependent code
						Camera = new Enki::CircularCam();
						m_EnkiCircularCamReader.Read(enkiIOStruct.SetNodePtr(subnodePtr), *Camera);
						if (enkiIOStruct.Good()) {
							Camera->setOwner(m_EnkiNeuralRobot.GetEnkiRobot());
							m_EnkiNeuralRobot.GetEnkiRobot()->cameras.push_back(Camera);
							m_EnkiNeuralRobot.GetEnkiRobot()->addLocalInteraction(Camera);
							readSuccess = true;
						}
						else
						{
							enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: unable to read camera");
							readSuccess = false;
							delete Camera;
							break;
						}
					}
					if (!readSuccess)
						break;
				}
				else
				{
					enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: subnodePtr->GetnodeType() != MSXML2::NODE_ELEMENT");
					readSuccess = false;
					break;
				}
			}
		} else { enkiIOStruct.Error("Error in CEnkiNeuralRobotIO: bad nodeName"); return; }
	}
}