//--------------------------------------------------------------------------------------
// File: RobotNeuralSimulator.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// 2D Simulator of simple robot controled by human (keyboard) or neural network
// This simulator is based on:
//
//  liban - base function to do usefull things related to simulation and scientific work
//  Copyright (C) 1999-2005 Stephane Magnenat <nct@ysagoon.com>
//  Copyright (C) 2004-2005 Antoine Beyeler <antoine.beyeler@epfl.ch>
//  Copyright (C) 2005 Laboratory of Intelligent Systems, EPFL, Lausanne
//
// and
//
//  Enki - a fast 2D robot simulator
//  Copyright (C) 1999-2005 Stephane Magnenat <nct@ysagoon.com>
//  Copyright (C) 2005 Laboratory of Intelligent Systems, EPFL, Lausanne
//
// This project was created by DirectX Sample Browser and was
// directly derived from SimpleSample
//
//  Copyright (C) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"
#include "resource.h"

#pragma warning(disable : 4530)

#include "RobotNeuralSimulator.h"

//#define DEBUG_VS   // Uncomment this line to debug vertex shaders 
//#define DEBUG_PS   // Uncomment this line to debug pixel shaders 

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
ID3DXFont*              g_pFont = NULL;         // Font for drawing text
ID3DXSprite*            g_pTextSprite = NULL;   // Sprite for batching draw text calls
bool                    g_bShowHelp = false;    // If true, it renders the UI control text
CDXUTDialogResourceManager g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg         g_SettingsDlg;          // Device settings dialog
CDXUTDialog             g_HUD;                  // dialog for standard controls
CDXUTDialog             g_SampleUI;             // dialog for sample specific controls
CDXUTLineManager        g_LineManager;          // line manager

CRobotNeuralSimulator   g_RobotNeuralSimulator;

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN    1
#define IDC_TOGGLEREF           2
#define IDC_CHANGEDEVICE        3

#define IDC_FILENAMESTATIC      10
#define IDC_FILENAMEEDIT        15
#define IDC_LOADSCENEBUTTON     20
#define IDC_LOADROBOTBUTTON     25
#define IDC_LOADNETWORKBUTTON   30
#define IDC_KEYBCONTROLRADIO    35
#define IDC_NEURCONTROLRADIO    40
#define IDC_NONETARGETRADIO     45
#define IDC_ROBOTTARGETRADIO    50
#define IDC_ZOOMSLIDER          55
#define IDC_RESETVIEWBUTTON     60
#define IDC_STARTLOGBUTTON      65
#define IDC_STOPLOGBUTTON       70

#define ID_TARGETGROUP          1
#define ID_REMOTECONTROLGROUP   2

//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
bool    CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext );
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void    CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void    CALLBACK OnLostDevice( void* pUserContext );
void    CALLBACK OnDestroyDevice( void* pUserContext );

void    InitApp();

void    OnLoadScene();
void    OnLoadRobot();
void    OnLoadNetwork();
void    OnStartLog();
void    OnStopLog();

float   GetZoomSlider();
void    SetZoomSlider();

void    RenderText();

void OnLoadScene()
{
	std::wstring wfileName = g_SampleUI.GetEditBox(IDC_FILENAMEEDIT)->GetText();
	std::string fileName(wfileName.begin(), wfileName.end());
	if (fileName == "")
	{
		g_RobotNeuralSimulator.GetScene()->ClearScene();
		g_RobotNeuralSimulator.SetStatus("Scene was cleared successfully");
	}
	else
	{
		g_RobotNeuralSimulator.GetScene()->LoadSceneFromFile(fileName);
		if (g_RobotNeuralSimulator.HandleError())
			g_RobotNeuralSimulator.SetStatus("Scene was loaded successfully");
	}
	g_RobotNeuralSimulator.ResetViewport();
	g_SampleUI.GetRadioButton(IDC_NONETARGETRADIO)->SetChecked(true);
	g_SampleUI.GetRadioButton(IDC_KEYBCONTROLRADIO)->SetChecked(true);
}

void OnLoadRobot()
{
	std::wstring wfileName = g_SampleUI.GetEditBox(IDC_FILENAMEEDIT)->GetText();
	std::string fileName(wfileName.begin(), wfileName.end());
	if (fileName == "")
	{
		g_RobotNeuralSimulator.GetScene()->ClearRobot();
		g_RobotNeuralSimulator.SetStatus("Robot was cleared successfully");
	}
	else
	{
		g_RobotNeuralSimulator.GetScene()->LoadRobotFromFile(fileName);
		g_RobotNeuralSimulator.GetScene()->SetRobotToKeyboard();
		if (g_RobotNeuralSimulator.HandleError())
			g_RobotNeuralSimulator.SetStatus("Robot was loaded successfully");
	}
	g_SampleUI.GetRadioButton(IDC_KEYBCONTROLRADIO)->SetChecked(true);
}

void OnLoadNetwork()
{
	std::wstring wfileName = g_SampleUI.GetEditBox(IDC_FILENAMEEDIT)->GetText();
	std::string fileName(wfileName.begin(), wfileName.end());
	if (fileName == "")
	{
		g_RobotNeuralSimulator.GetScene()->ClearNetwork();
		g_RobotNeuralSimulator.SetStatus("Neural network was cleared successfully");
	}
	else
	{
		g_RobotNeuralSimulator.GetScene()->LoadNetworkFromFile(fileName);
		if (g_RobotNeuralSimulator.HandleError())
			g_RobotNeuralSimulator.SetStatus("Neural network was loaded successfully");
	}
	g_RobotNeuralSimulator.GetScene()->SetRobotToKeyboard();
	g_SampleUI.GetRadioButton(IDC_KEYBCONTROLRADIO)->SetChecked(true);
}

void OnStartLog()
{
	std::wstring wfileName = g_SampleUI.GetEditBox(IDC_FILENAMEEDIT)->GetText();
	std::string fileName(wfileName.begin(), wfileName.end());
	if (fileName == "")
	{
		g_RobotNeuralSimulator.SetStatus("Please enter the name of the log");
	}
	else
	{
		g_RobotNeuralSimulator.GetScene()->StartLog(fileName);
		if (g_RobotNeuralSimulator.HandleError())
			g_RobotNeuralSimulator.SetStatus("Log was started successfully");
	}
}

void OnStopLog()
{
	bool wasLogging = g_RobotNeuralSimulator.GetScene()->isLogging();
	g_RobotNeuralSimulator.GetScene()->StopLog();
	if (g_RobotNeuralSimulator.HandleError() && wasLogging)
		g_RobotNeuralSimulator.SetStatus("Log was stopped successfully");
}

float GetZoomSlider()
{
	float fScaleMin = 0.01f;
	float fScaleMax = 0.05f;
	float fPos = (float)g_SampleUI.GetSlider(IDC_ZOOMSLIDER)->GetValue() / 100.0f;
	return fScaleMin + fPos * (fScaleMax - fScaleMin);
}

void SetZoomSlider()
{
	float fScale = g_RobotNeuralSimulator.GetScaling();
	float fScaleMin = 0.01f;
	float fScaleMax = 0.05f;
	float fPos = (fScale - fScaleMin) / (fScaleMax - fScaleMin);
	if (fPos < 0.0f) fPos = 0.0f;
	if (fPos > 1.0f) fPos = 1.0f;
	g_SampleUI.GetSlider(IDC_ZOOMSLIDER)->SetValue((int)(100.0f * fPos));
}


//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
	CoInitialize(NULL);

    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // Set the callback functions. These functions allow DXUT to notify
    // the application about device changes, user input, and windows messages.  The 
    // callbacks are optional so you need only set callbacks for events you're interested 
    // in. However, if you don't handle the device reset/lost callbacks then the sample 
    // framework won't be able to reset your device since the application must first 
    // release all device resources before resetting.  Likewise, if you don't handle the 
    // device created/destroyed callbacks then DXUT won't be able to 
    // recreate your device resources.
    DXUTSetCallbackDeviceCreated( OnCreateDevice );
    DXUTSetCallbackDeviceReset( OnResetDevice );
    DXUTSetCallbackDeviceLost( OnLostDevice );
    DXUTSetCallbackDeviceDestroyed( OnDestroyDevice );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( KeyboardProc );
    DXUTSetCallbackFrameRender( OnFrameRender );
    DXUTSetCallbackFrameMove( OnFrameMove );

    // Show the cursor and clip it when in full screen
    DXUTSetCursorSettings( true, true );

    InitApp();

    // Initialize DXUT and create the desired Win32 window and Direct3D 
    // device for the application. Calling each of these functions is optional, but they
    // allow you to set several options which control the behavior of the framework.
    DXUTInit( true, true, true ); // Parse the command line, handle the default hotkeys, and show msgboxes
    DXUTCreateWindow( L"Robot Neural Simulator" );
    DXUTCreateDevice( D3DADAPTER_DEFAULT, true, 640, 480, IsDeviceAcceptable, ModifyDeviceSettings );

    // Pass control to DXUT for handling the message pump and 
    // dispatching render calls. DXUT will call your FrameMove 
    // and FrameRender callback when there is idle time between handling window messages.
    DXUTMainLoop();

    // Perform any application-level cleanup here. Direct3D device resources are released within the
    // appropriate callback functions and therefore don't require any cleanup code here.

    return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
	// Initialize dialogs
	g_SettingsDlg.Init( &g_DialogResourceManager );
	g_HUD.Init( &g_DialogResourceManager );
	g_SampleUI.Init( &g_DialogResourceManager );

	g_HUD.SetCallback( OnGUIEvent ); int iY = 10; 
	g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
	g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22 );
	g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );

	g_SampleUI.SetCallback( OnGUIEvent ); iY = 10; 

	//    TODO: add UI controls as needed
	g_SampleUI.AddStatic(IDC_FILENAMESTATIC, L"File Name:", 35, iY += 24, 125, 22);
	g_SampleUI.AddEditBox(IDC_FILENAMEEDIT, L"Default.xml", 35, iY += 24, 125, 32);
	g_SampleUI.AddButton(IDC_LOADSCENEBUTTON, L"Load Scene", 35, iY += 44, 125, 22);
	g_SampleUI.AddButton(IDC_LOADROBOTBUTTON, L"Load Robot", 35, iY += 24, 125, 22);
	g_SampleUI.AddButton(IDC_LOADNETWORKBUTTON, L"Load Network", 35, iY += 24, 125, 22);
	g_SampleUI.AddRadioButton(IDC_KEYBCONTROLRADIO, ID_REMOTECONTROLGROUP, L"Keyboard", 35, iY += 24, 125, 22);
	g_SampleUI.AddRadioButton(IDC_NEURCONTROLRADIO, ID_REMOTECONTROLGROUP, L"Neural Network", 35, iY += 24, 125, 22);
	g_SampleUI.AddRadioButton(IDC_NONETARGETRADIO, ID_TARGETGROUP, L"No Target", 35, iY += 24, 125, 22);
	g_SampleUI.AddRadioButton(IDC_ROBOTTARGETRADIO, ID_TARGETGROUP, L"Robot Target", 35, iY += 24, 125, 22);
	g_SampleUI.AddButton(IDC_RESETVIEWBUTTON, L"Reset Zoom", 35, iY += 24, 125, 22);
	g_SampleUI.AddSlider(IDC_ZOOMSLIDER, 50, iY += 24, 100, 22);
	g_SampleUI.AddButton(IDC_STARTLOGBUTTON, L"Start Log", 35, iY += 24, 125, 22);
	g_SampleUI.AddButton(IDC_STOPLOGBUTTON, L"Stop Log", 35, iY += 24, 125, 22);
	g_SampleUI.GetRadioButton(IDC_KEYBCONTROLRADIO)->SetChecked(true);
	g_SampleUI.GetRadioButton(IDC_NONETARGETRADIO)->SetChecked(true);

	g_RobotNeuralSimulator.ResetViewport();
}


//--------------------------------------------------------------------------------------
// Called during device initialization, this code checks the device for some 
// minimum set of capabilities, and rejects those that don't pass by returning false.
//--------------------------------------------------------------------------------------
bool CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, 
                                  D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    // Skip backbuffer formats that don't support alpha blending
    IDirect3D9* pD3D = DXUTGetD3DObject(); 
    if( FAILED( pD3D->CheckDeviceFormat( pCaps->AdapterOrdinal, pCaps->DeviceType,
                    AdapterFormat, D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING, 
                    D3DRTYPE_TEXTURE, BackBufferFormat ) ) )
        return false;

    return true;
}


//--------------------------------------------------------------------------------------
// This callback function is called immediately before a device is created to allow the 
// application to modify the device settings. The supplied pDeviceSettings parameter 
// contains the settings that the framework has selected for the new device, and the 
// application can make any desired changes directly to this structure.  Note however that 
// DXUT will not correct invalid device settings so care must be taken 
// to return valid device settings, otherwise IDirect3D9::CreateDevice() will fail.  
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext )
{
	// If device doesn't support HW T&L or doesn't support 1.1 vertex shaders in HW 
	// then switch to SWVP.
	if( (pCaps->DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) == 0 ||
		 pCaps->VertexShaderVersion < D3DVS_VERSION(1,1) )
	{
		pDeviceSettings->BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	}

	// Debugging vertex shaders requires either REF or software vertex processing 
	// and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
	if( pDeviceSettings->DeviceType != D3DDEVTYPE_REF )
	{
		pDeviceSettings->BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
		pDeviceSettings->BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
		pDeviceSettings->BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	}
#endif
#ifdef DEBUG_PS
	pDeviceSettings->DeviceType = D3DDEVTYPE_REF;
#endif

	// For the first device created if its a REF device, optionally display a warning dialog box
	static bool s_bFirstTime = true;
	if( s_bFirstTime )
	{
		s_bFirstTime = false;
		if( pDeviceSettings->DeviceType == D3DDEVTYPE_REF )
			DXUTDisplaySwitchingToREFWarning();
	}

	return true;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// created, which will happen during application initialization and windowed/full screen 
// toggles. This is the best location to create D3DPOOL_MANAGED resources since these 
// resources need to be reloaded whenever the device is destroyed. Resources created  
// here should be released in the OnDestroyDevice callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	HRESULT hr;

	V_RETURN( g_DialogResourceManager.OnCreateDevice( pd3dDevice ) );
	V_RETURN( g_SettingsDlg.OnCreateDevice( pd3dDevice ) );

	// Initialize the font
	V_RETURN( D3DXCreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
						 OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
						 L"Verdana", &g_pFont ) );

	V_RETURN( g_LineManager.OnCreatedDevice( pd3dDevice ) );

	V_RETURN( g_RobotNeuralSimulator.OnCreatedDevice( pd3dDevice ) );

	// Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the 
	// shader debugger. Debugging vertex shaders requires either REF or software vertex 
	// processing, and debugging pixel shaders requires REF.  The 
	// D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug experience in the 
	// shader debugger.  It enables source level debugging, prevents instruction 
	// reordering, prevents dead code elimination, and forces the compiler to compile 
	// against the next higher available software target, which ensures that the 
	// unoptimized shaders do not exceed the shader model limitations.  Setting these 
	// flags will cause slower rendering since the shaders will be unoptimized and 
	// forced into software.  See the DirectX documentation for more information about 
	// using the shader debugger.
	DWORD dwShaderFlags = D3DXFX_NOT_CLONEABLE;

	#if defined( DEBUG ) || defined( _DEBUG )
	// Set the D3DXSHADER_DEBUG flag to embed debug information in the shaders.
	// Setting this flag improves the shader debugging experience, but still allows 
	// the shaders to be optimized and to run exactly the way they will run in 
	// the release configuration of this program.
	dwShaderFlags |= D3DXSHADER_DEBUG;
	#endif

	#ifdef DEBUG_VS
		dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
	#endif
	#ifdef DEBUG_PS
		dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
	#endif

	return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has been 
// reset, which will happen after a lost device scenario. This is the best location to 
// create D3DPOOL_DEFAULT resources since these resources need to be reloaded whenever 
// the device is lost. Resources created here should be released in the OnLostDevice 
// callback. 
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, 
                                const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
	HRESULT hr;

	V_RETURN( g_DialogResourceManager.OnResetDevice() );
	V_RETURN( g_SettingsDlg.OnResetDevice() );

	if( g_pFont )
		V_RETURN( g_pFont->OnResetDevice() );

	// Create a sprite to help batch calls when drawing many lines of text
	V_RETURN( D3DXCreateSprite( pd3dDevice, &g_pTextSprite ) );

	V_RETURN( g_LineManager.OnResetDevice() );

	V_RETURN( g_RobotNeuralSimulator.OnResetDevice() );

	// Turn off culling, so we see the front and back of the triangle
	pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

	// Turn off D3D lighting, since we are providing our own vertex colors
	pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );

	int x1 = 10;
	int y1 = 55;
	int x2 = pBackBufferSurfaceDesc->Width - 170;
	int y2 = pBackBufferSurfaceDesc->Height - 35;

	if (x2 <= x1 || y2 <= y1)
	{
		x1 = 0;
		y1 = 0;
		x2 = pBackBufferSurfaceDesc->Width;
		y2 = pBackBufferSurfaceDesc->Height;
	}

	RECT rc;
	rc.left = x1;
	rc.top  = y1;
	rc.right = x2;
	rc.bottom = y2;

	g_LineManager.RemoveAllLines();
	g_LineManager.AddRect(0, rc, D3DCOLOR_RGBA(0,0,127,255), 2.0f, 1.0f, true);

	int X = x1;
	int Y = y1;
	int Width = x2 - x1;
	int Height = y2 - y1;

	g_RobotNeuralSimulator.SetClippingRect(X, Y, Width, Height);

	g_HUD.SetLocation( pBackBufferSurfaceDesc->Width-170, 0 );
	g_HUD.SetSize( 170, 170 );
	g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-170, pBackBufferSurfaceDesc->Height-400 );
	g_SampleUI.SetSize( 170, 240 );

	return S_OK;
}


//--------------------------------------------------------------------------------------
// This callback function will be called once at the beginning of every frame. This is the
// best location for your application to handle updates to the scene, but is not 
// intended to contain actual rendering calls, which should instead be placed in the 
// OnFrameRender callback.  
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
	g_RobotNeuralSimulator.OnFrameMove(fTime, fElapsedTime);
	g_RobotNeuralSimulator.ZoomViewportByKeys(fElapsedTime);
	SetZoomSlider();
	if (g_SampleUI.GetRadioButton(IDC_NONETARGETRADIO)->GetChecked())
	{
		g_RobotNeuralSimulator.MoveViewportByKeys(fElapsedTime);
	}
	else if (g_SampleUI.GetRadioButton(IDC_ROBOTTARGETRADIO)->GetChecked())
	{
		g_RobotNeuralSimulator.MoveViewportToRobot(fElapsedTime);
	}
}


//--------------------------------------------------------------------------------------
// This callback function will be called at the end of every frame to perform all the 
// rendering calls for the scene, and it will also be called if the window needs to be 
// repainted. After this function has returned, DXUT will call 
// IDirect3DDevice9::Present to display the contents of the next buffer in the swap chain
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
	HRESULT hr;

	// If the settings dialog is being shown, then
	// render it instead of rendering the app's scene
	if( g_SettingsDlg.IsActive() )
	{
		g_SettingsDlg.OnRender( fElapsedTime );
		return;
	}

	// Clear the render target and the zbuffer 
	V( pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_ARGB(0, 45, 50, 170), 1.0f, 0) );

	// Render the scene
	if( SUCCEEDED( pd3dDevice->BeginScene() ) )
	{
		g_RobotNeuralSimulator.OnFrameRender(fTime, fElapsedTime);

		DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" ); // These events are to help PIX identify what the code is doing		
		V( g_LineManager.OnRender() );
		RenderText();
		V( g_HUD.OnRender( fElapsedTime ) );
		V( g_SampleUI.OnRender( fElapsedTime ) );
		DXUT_EndPerfEvent();

		V( pd3dDevice->EndScene() );
	}
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText()
{
	// The helper object simply helps keep track of text position, and color
	// and then it calls pFont->DrawText( m_pSprite, strMsg, -1, &rc, DT_NOCLIP, m_clr );
	// If NULL is passed in as the sprite object, then it will work however the 
	// pFont->DrawText() will not be batched together.  Batching calls will improves performance.
	CDXUTTextHelper txtHelper( g_pFont, g_pTextSprite, 15 );

	// Output statistics
	txtHelper.Begin();
	txtHelper.SetInsertionPos( 5, 5 );
	txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
	txtHelper.DrawTextLine( DXUTGetFrameStats(true) );
	txtHelper.DrawTextLine( DXUTGetDeviceStats() );
	
	//    TODO: add UI text as needed
	txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
	txtHelper.DrawTextLine( g_RobotNeuralSimulator.GetStatus().c_str() );

	if (g_RobotNeuralSimulator.GetScene()->isLogging())
	{
		txtHelper.SetInsertionPos( 20, 60 );
		txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 0.0f, 0.0f, 1.0f ) );
		txtHelper.DrawTextLine( L"REC" );
	}

	// Draw help
	const D3DSURFACE_DESC* pd3dsdBackBuffer = DXUTGetBackBufferSurfaceDesc();
	if( g_bShowHelp )
	{
		txtHelper.SetInsertionPos( 20, pd3dsdBackBuffer->Height-20*6 );
		txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
		txtHelper.DrawTextLine( L"Controls (F1 to hide):" );
		txtHelper.DrawTextLine( L"  Arrows: to move the robot" );
		txtHelper.DrawTextLine( L"  W,S,A,D: to move the camera" );
		txtHelper.DrawTextLine( L"  Q,E: zoom" );
		txtHelper.DrawTextLine( L"  ESC: quit" );
	}
	else
	{
		txtHelper.SetInsertionPos( 10, pd3dsdBackBuffer->Height-15*2 );
		txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
		txtHelper.DrawTextLine( L"Press F1 for help" );
	}

	txtHelper.End();
}


//--------------------------------------------------------------------------------------
// Before handling window messages, DXUT passes incoming windows 
// messages to the application through this callback function. If the application sets 
// *pbNoFurtherProcessing to TRUE, then DXUT will not process this message.
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext )
{
	// Always allow dialog resource manager calls to handle global messages
	// so GUI state is updated correctly
	*pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;

	if( g_SettingsDlg.IsActive() )
	{
		g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
		return 0;
	}

	// Give the dialogs a chance to handle the message first
	*pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;
	*pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
	if( *pbNoFurtherProcessing )
		return 0;

	g_RobotNeuralSimulator.MsgProc( hWnd, uMsg, wParam, lParam );

	return 0;
}


//--------------------------------------------------------------------------------------
// As a convenience, DXUT inspects the incoming windows messages for
// keystroke messages and decodes the message parameters to pass relevant keyboard
// messages to the application.  The framework does not remove the underlying keystroke 
// messages, which are still passed to the application's MsgProc callback.
//--------------------------------------------------------------------------------------
void CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	if( bKeyDown )
	{
		switch( nChar )
		{
			case VK_F1: g_bShowHelp = !g_bShowHelp; break;
		}
	}
}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
	switch( nControlID )
	{
		case IDC_TOGGLEFULLSCREEN:  DXUTToggleFullScreen(); break;
		case IDC_TOGGLEREF:         DXUTToggleREF(); break;
		case IDC_CHANGEDEVICE:      g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;

		case IDC_LOADSCENEBUTTON:   OnLoadScene(); break;
		case IDC_LOADROBOTBUTTON:   OnLoadRobot(); break;
		case IDC_LOADNETWORKBUTTON: OnLoadNetwork(); break;
		case IDC_KEYBCONTROLRADIO:  g_RobotNeuralSimulator.GetScene()->SetRobotToKeyboard();      g_RobotNeuralSimulator.HandleError(); break;
		case IDC_NEURCONTROLRADIO:  g_RobotNeuralSimulator.GetScene()->SetRobotToNeuralNetwork(); g_RobotNeuralSimulator.HandleError(); break;
		case IDC_ZOOMSLIDER:        g_RobotNeuralSimulator.SetScaling(GetZoomSlider()); break;
		case IDC_RESETVIEWBUTTON:   g_RobotNeuralSimulator.ResetViewport(); g_SampleUI.GetRadioButton(IDC_NONETARGETRADIO)->SetChecked(true); break;
		case IDC_STARTLOGBUTTON:    OnStartLog(); break;
		case IDC_STOPLOGBUTTON:     OnStopLog(); break;
	}
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// entered a lost state and before IDirect3DDevice9::Reset is called. Resources created
// in the OnResetDevice callback should be released here, which generally includes all 
// D3DPOOL_DEFAULT resources. See the "Lost Devices" section of the documentation for 
// information about lost devices.
//--------------------------------------------------------------------------------------
void CALLBACK OnLostDevice( void* pUserContext )
{
	g_DialogResourceManager.OnLostDevice();
	g_SettingsDlg.OnLostDevice();
	g_LineManager.OnLostDevice();
	g_RobotNeuralSimulator.OnLostDevice();
	if( g_pFont )
		g_pFont->OnLostDevice();
	SAFE_RELEASE( g_pTextSprite );
}


//--------------------------------------------------------------------------------------
// This callback function will be called immediately after the Direct3D device has 
// been destroyed, which generally happens as a result of application termination or 
// windowed/full screen toggles. Resources created in the OnCreateDevice callback 
// should be released here, which generally includes all D3DPOOL_MANAGED resources. 
//--------------------------------------------------------------------------------------
void CALLBACK OnDestroyDevice( void* pUserContext )
{
	g_DialogResourceManager.OnDestroyDevice();
	g_SettingsDlg.OnDestroyDevice();
	g_LineManager.OnDeletedDevice();
	g_RobotNeuralSimulator.OnDestroyDevice();
	SAFE_RELEASE( g_pFont );
}