#ifndef ROBOTNEURALSIMULATOR_H
#define ROBOTNEURALSIMULATOR_H

//--------------------------------------------------------------------------------------
// File: RobotNeuralSimulator.h
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Neural Simulator header
// This is the basic header for the application
//
//--------------------------------------------------------------------------------------

#include <string>
#include <sstream>
#include <iomanip>

#include "MsgObserverPattern.h"
#include "D3DDeviceObserverPattern.h"

#include "MsgSubject.h"
#include "D3DDeviceSubject.h"

#include "enki/PhysicalEngine.h"
#include "enki/robots/NeuralRobot.h"

#include "RFWCoreClasses.h"
#include "RFWNeuralRobot.h"

#include "CustomVertex.h"
#include "VertexBuffer.h"
#include "Keyboard.h"
#include "Viewport.h"
#include "StatusBar.h"
#include "PositioningSystem.h"

#include "RobotNeuralSimulatorScene.h"

//--------------------------------------------------------------------------------------
// CRobotNeuralSimulatorScene class
//--------------------------------------------------------------------------------------
class CRobotNeuralSimulator
{
protected:
	CRobotNeuralSimulator(const CRobotNeuralSimulator &);
	const CRobotNeuralSimulator & operator=(const CRobotNeuralSimulator &);
public:
	CRobotNeuralSimulator():
	  m_PositioningSystem(m_Viewport),
	  m_Scene(m_VertexBuffer, m_Keyboard)
	{
		m_MsgSubject.RegisterObserver(&m_Keyboard);
		m_MsgSubject.RegisterObserver(&m_PositioningSystem);
		m_D3DDeviceSubject.RegisterObserver(&m_VertexBuffer);
		m_D3DDeviceSubject.RegisterObserver(&m_Viewport);
	}
	virtual ~CRobotNeuralSimulator() { }

	void MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		m_MsgSubject.NotifyObservers(hWnd, uMsg, wParam, lParam);
	}

	HRESULT OnCreatedDevice( IDirect3DDevice9* pd3dDevice ) { return m_D3DDeviceSubject.NotifyCreatedDevice(pd3dDevice); }
	HRESULT OnResetDevice()   { return m_D3DDeviceSubject.NotifyResetDevice(); }
	HRESULT OnLostDevice()    { return m_D3DDeviceSubject.NotifyLostDevice(); }
	HRESULT OnDestroyDevice() { return m_D3DDeviceSubject.NotifyDestroyDevice(); }

	bool HandleError();

	std::wstring GetStatus() { return m_StatusBar.GetText(); }
	void SetStatus(std::string text) { m_StatusBar.SetNormalPriorityText(text); }

	CKeyboard * GetKeyboard() { return &m_Keyboard; }
	CViewport * GetViewport() { return &m_Viewport; }
	CRobotNeuralSimulatorScene * GetScene() { return &m_Scene; }

	float GetScaling() { return m_Viewport.GetScaling(); }
	void  SetScaling(float fScaling) { m_Viewport.SetScaling(fScaling); }
	void  SetClippingRect(int X, int Y, int Width, int Height) { m_Viewport.SetClippingRect(X, Y, Width, Height); }

	void ResetViewport();
	void ZoomViewportByKeys(float fElapsedTime);
	void MoveViewportByKeys(float fElapsedTime);
	void MoveViewportToRobot(float fElapsedTime);

	void OnFrameMove(double fTime, float fElapsedTime);
	void OnFrameRender(double fTime, float fElapsedTime);
protected:
	void SetStatusCoordinates();

	CMsgSubject        m_MsgSubject;
	CD3DDeviceSubject  m_D3DDeviceSubject;
	CVertexBuffer      m_VertexBuffer;
	CKeyboard          m_Keyboard;
	CViewport          m_Viewport;
	CStatusBar         m_StatusBar;
	CPositioningSystem m_PositioningSystem;
	CRobotNeuralSimulatorScene m_Scene;
};

bool CRobotNeuralSimulator::HandleError()
{
	if (!m_Scene.Good())
	{
		m_StatusBar.SetHighPriorityText(m_Scene.ReportError());
		m_Scene.ClearError();
		return false;
	}
	else return true;
}

void CRobotNeuralSimulator::ResetViewport()
{
	m_Viewport.SetToTarget(
		(float)m_Scene.GetWorld()->GetEnkiWorld()->w,
		(float)m_Scene.GetWorld()->GetEnkiWorld()->h);
}

void CRobotNeuralSimulator::ZoomViewportByKeys(float fElapsedTime)
{
	if      (m_Keyboard.isKey('E')) { m_Viewport.ScaleBy(+2.0f, fElapsedTime); }
	else if (m_Keyboard.isKey('Q')) { m_Viewport.ScaleBy(+0.5f, fElapsedTime); }
}

void CRobotNeuralSimulator::MoveViewportByKeys(float fElapsedTime)
{
	if      (m_Keyboard.isKey('W')) m_Viewport.MoveBy(0.0f, +20.0f, fElapsedTime);
	else if (m_Keyboard.isKey('S')) m_Viewport.MoveBy(0.0f, -20.0f, fElapsedTime);
	if      (m_Keyboard.isKey('D')) m_Viewport.MoveBy(+20.0f, 0.0f, fElapsedTime);
	else if (m_Keyboard.isKey('A')) m_Viewport.MoveBy(-20.0f, 0.0f, fElapsedTime);
}

void CRobotNeuralSimulator::MoveViewportToRobot(float fElapsedTime)
{
	if (m_Scene.GetRobot())
	{
		m_Viewport.MoveTo(
			(float)m_Scene.GetRobot()->GetEnkiRobot()->pos.x,
			(float)m_Scene.GetRobot()->GetEnkiRobot()->pos.y,
			fElapsedTime);
	}
}

void CRobotNeuralSimulator::SetStatusCoordinates()
{
	std::wstringstream wstream;
	if (m_PositioningSystem.IsActualDefined())
	{
		D3DXVECTOR2 localActual = m_PositioningSystem.GetLocalActual();
		wstream << std::setw(5);
		wstream << localActual.x << L" " << localActual.y;
		m_StatusBar.SetText(wstream.str(), 0.1f, CStatusBar::Low);
	}
}

void CRobotNeuralSimulator::OnFrameMove(double fTime, float fElapsedTime)
{
	m_StatusBar.SetActualTime((float)fTime);
	SetStatusCoordinates();

	m_Scene.Step(fElapsedTime);
	if (m_Scene.isLogging())
		m_Scene.Log(fElapsedTime);
}

void CRobotNeuralSimulator::OnFrameRender(double fTime, float fElapsedTime)
{
	HRESULT hr;

	V( m_Viewport.SetViewport() );
	V( m_Viewport.ClearViewport(D3DCOLOR_ARGB(0, 0, 0, 170)) );
	m_Scene.Render();
	V( m_Viewport.UndoViewport() );
}

#endif // ROBOTNEURALSIMULATOR_H