//--------------------------------------------------------------------------------------
// File: RobotNeuralSimulatorScene.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Robot Neural Simulator scene source
// CRobotNeuralSimulatorScene class encapsulates all the basic functionality
// of the simulator
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning(disable : 4530)

#include "enki/PhysicalEngine.h"
#include "enki/robots/NeuralRobot.h"

#include "VertexBuffer.h"
#include "Keyboard.h"

#include "Log.h"
#include "RemoteControl.h"

#include "RFWCoreClasses.h"
#include "RFWCoreIO.h"
#include "RFWNeuralNetwork.h"
#include "RFWNeuralNetworkIO.h"
#include "RFWNeuralRobot.h"

#include "RobotNeuralSimulatorScene.h"

//--------------------------------------------------------------------------------------
// CRobotNeuralSimulatorScene class
//--------------------------------------------------------------------------------------

const double CRobotNeuralSimulatorScene::defaultWidth  = 100.0;
const double CRobotNeuralSimulatorScene::defaultHeight = 100.0;

CRobotNeuralSimulatorScene::CRobotNeuralSimulatorScene(CVertexBuffer &vertexBuffer, CKeyboard &keyboard):
	m_VertexBuffer(vertexBuffer),
	m_Keyboard(keyboard),
	m_RobotToNeuralNetwork(false),
	m_KeyboardRemoteControl(0),
	m_NeuralNetworkRemoteControl(0),
	m_EnkiWorld(0),
	m_EnkiNeuralRobot(0)
{
	m_Log = new CLog();
	m_KeyboardRemoteControl = new CKeyboardRemoteControl(keyboard);
	m_NeuralNetworkRemoteControl = new CNeuralNetworkRemoteControl();
	m_EnkiWorld = new RFW::CEnkiWorld(vertexBuffer);
	SetRobotToKeyboard();
}

CRobotNeuralSimulatorScene::~CRobotNeuralSimulatorScene()
{
	ClearScene();
	delete m_Log;
	delete m_KeyboardRemoteControl;
	delete m_NeuralNetworkRemoteControl;
	delete m_EnkiWorld;
}

void CRobotNeuralSimulatorScene::ClearError()
{
	m_Log->ClearError();
	m_ReportError.ClearError();
}

void CRobotNeuralSimulatorScene::ClearScene()
{
	m_EnkiWorld->ClearWorld(defaultWidth, defaultHeight);
	m_EnkiNeuralRobot = 0;
}

void CRobotNeuralSimulatorScene::ClearRobot()
{
	if (m_EnkiNeuralRobot)
	{
		m_EnkiWorld->RemoveRobot(m_EnkiNeuralRobot);
		m_EnkiNeuralRobot = 0;
	}
}

void CRobotNeuralSimulatorScene::ClearNetwork()
{
	m_NeuralNetworkRemoteControl->Clear();
}

bool CRobotNeuralSimulatorScene::LoadSceneFromFile(std::string fileName)
{
	RFW::CEnkiIOStruct enkiIOStruct(fileName);
	if (enkiIOStruct.Good())
	{
		ClearScene();
		m_EnkiWorld->Read(enkiIOStruct);
		if (enkiIOStruct.Good())
			return true;
		else
		{
			ClearScene();
			Error(enkiIOStruct.ReportError());
			return false;
		}
	}
	else
	{
		Error(enkiIOStruct.ReportError());
		return false;
	}
}

bool CRobotNeuralSimulatorScene::LoadRobotFromFile(std::string fileName)
{
	RFW::CEnkiIOStruct enkiIOStruct(fileName);
	if (enkiIOStruct.Good())
	{
		ClearRobot();
		m_EnkiNeuralRobot = new RFW::CEnkiNeuralRobot((RFW::CEnkiWorld &)*m_EnkiWorld);
		m_EnkiNeuralRobot->Read(enkiIOStruct);
		if (enkiIOStruct.Good())
		{
			if (m_EnkiWorld->GetStartingPointCount() > 0)
				m_EnkiNeuralRobot->GetEnkiRobot()->pos = m_EnkiWorld->GetStartingPoint(0);
			return true;
		}
		else
		{
			ClearRobot();
			Error(enkiIOStruct.ReportError());
			return false;
		}
	}
	else
	{
		Error(enkiIOStruct.ReportError());
		return false;
	}
}

bool CRobotNeuralSimulatorScene::LoadNetworkFromFile(std::string fileName)
{
	RFW::CEnkiIOStruct enkiIOStruct(fileName);
	if (enkiIOStruct.Good())
	{
		RFW::CFeedForwardReader feedForwardReader;
		m_NeuralNetworkRemoteControl->Clear();
		feedForwardReader.Read(enkiIOStruct, m_NeuralNetworkRemoteControl->GetNeuralNetwork());
		if (enkiIOStruct.Good())
			return true;
		else
		{
			Error(enkiIOStruct.ReportError());
			return false;
		}
	}
	else
	{
		Error(enkiIOStruct.ReportError());
		return false;
	}
}

bool CRobotNeuralSimulatorScene::StartLog(std::string fileName)
{
	if (!GetRobot())
	{
		Error("Error in CRobotNeuralSimulatorScene: robot is not loaded");
		return false;
	}
	m_Log->OpenLog(fileName);
	m_Log->Attach(GetRobot()->GetEnkiRobot());
	m_Log->StartLog();
	if (!m_Log->Good())
	{
		Error(m_Log->ReportError());
		Error("Error in CRobotNeuralSimulatorScene: unable to start log");
		m_Log->CloseLog();
		return false;
	}
	return true;
}

bool CRobotNeuralSimulatorScene::isLogging()
{
	return m_Log->isLogging();
}

void CRobotNeuralSimulatorScene::Log(float fElapsedTime)
{
	m_Log->Log(fElapsedTime);
}

bool CRobotNeuralSimulatorScene::StopLog()
{
	m_Log->CloseLog();
	return true;
}

bool CRobotNeuralSimulatorScene::SetRobotToKeyboard()
{
	if (!GetRobot()) return false;
	GetRobot()->GetEnkiRobot()->SetRemoteControl(m_KeyboardRemoteControl);
	m_RobotToNeuralNetwork = false;
	return true;
}

bool CRobotNeuralSimulatorScene::SetRobotToNeuralNetwork()
{
	if (!GetRobot()) return false;
	GetRobot()->GetEnkiRobot()->SetRemoteControl(m_NeuralNetworkRemoteControl);
	if (!m_NeuralNetworkRemoteControl->IsConsistent())
	{
		SetRobotToKeyboard();
		Error("Inconsistent neural network");
		return false;
	}
	else
	{
		m_RobotToNeuralNetwork = true;
		return true;
	}
}

void CRobotNeuralSimulatorScene::Render()
{
	m_EnkiWorld->Render();
}

void CRobotNeuralSimulatorScene::Step(double time)
{
	m_EnkiWorld->Step(time);
}