//--------------------------------------------------------------------------------------
// File: VertexBuffer.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Vertex buffer for DirectX 9.0 source
// Implementation of basic classes for managing vertex buffer
//
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning(disable : 4530)

#include "CustomVertex.h"
#include "VertexBuffer.h"

//--------------------------------------------------------------------------------------
// CVertexBuffer class implements ID3DDeviceObserver interface
//--------------------------------------------------------------------------------------
CVertexBuffer::CVertexBuffer():
  m_Size(512*sizeof(CUSTOMVERTEX)),
  m_LockedSize(0),
  m_pVB(0), m_pd3dDevice(0)
{

}

CVertexBuffer::~CVertexBuffer()
{

}

HRESULT CVertexBuffer::OnCreatedDevice(IDirect3DDevice9 *pd3dDevice)
{
	m_pd3dDevice = pd3dDevice;
	return S_OK;
}

HRESULT CVertexBuffer::OnResetDevice()
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	V_RETURN( m_pd3dDevice->CreateVertexBuffer(m_Size, 
		0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_pVB, NULL ) );
	m_LockedSize = 0;
	return S_OK;
}

HRESULT CVertexBuffer::OnLostDevice()
{
	SAFE_RELEASE( m_pVB );
	m_LockedSize = 0;
	return S_OK;
}

HRESULT CVertexBuffer::OnDestroyDevice()
{
	SAFE_RELEASE( m_pVB );
	m_LockedSize = 0;
	return S_OK;
}

HRESULT CVertexBuffer::Clear()
{
	m_LockedSize = 0;
	return S_OK;
}

HRESULT CVertexBuffer::CopyVertices(void *pData, UINT Size)
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	VOID* pVertices;
	if (Size >= m_Size)
	{
		SAFE_RELEASE( m_pVB );
		m_Size = 2*Size + 1;
		V_RETURN( m_pd3dDevice->CreateVertexBuffer(m_Size, 
			0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_pVB, NULL ) );
	}
	V_RETURN( m_pVB->Lock( 0, Size, &pVertices, 0 ) );
	memcpy( pVertices, pData, Size );
	m_pVB->Unlock();
	m_LockedSize = Size;
	return S_OK;
}

HRESULT CVertexBuffer::AddVertices(void *pData, UINT Size)
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	VOID* pVertices;
	if (m_LockedSize + Size >= m_Size)
	{
		SAFE_RELEASE( m_pVB );
		m_Size = 2*(m_LockedSize + Size) + 1;
		V_RETURN( m_pd3dDevice->CreateVertexBuffer(m_Size, 
			0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &m_pVB, NULL ) );
	}
	V_RETURN( m_pVB->Lock( m_LockedSize, Size, &pVertices, 0 ) );
	memcpy( pVertices, pData, Size );
	m_pVB->Unlock();
	m_LockedSize += Size;
	return S_OK;
}

HRESULT CVertexBuffer::SetVertexBuffer()
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	V_RETURN( m_pd3dDevice->SetStreamSource( 0, m_pVB, 0, sizeof(CUSTOMVERTEX) ) );
	V_RETURN( m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX ) );
	return S_OK;
}

HRESULT CVertexBuffer::DrawPrimitive(D3DPRIMITIVETYPE PrimitiveType)
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	int Total = m_LockedSize / sizeof(CUSTOMVERTEX);
	int Count = 0;
	switch (PrimitiveType)
	{
	case D3DPT_POINTLIST: Count = Total; break;
	case D3DPT_LINELIST : Count = Total/2; break;
	case D3DPT_LINESTRIP: Count = Total - 1; break;
	case D3DPT_TRIANGLELIST: Count = Total/3; break;
	case D3DPT_TRIANGLESTRIP: Count = Total - 2; break;
	case D3DPT_TRIANGLEFAN: Count = Total - 2; break;
	}
	if (Count > 0)
		V_RETURN( m_pd3dDevice->DrawPrimitive( PrimitiveType, 0, Count) );
	return S_OK;
}

HRESULT CVertexBuffer::DrawPrimitive(D3DPRIMITIVETYPE PrimitiveType, D3DXMATRIX & matTransform)
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;
	D3DXMATRIX oldWorld;
	D3DXMATRIX newWorld;
	V_RETURN( m_pd3dDevice->GetTransform( D3DTS_WORLD, &oldWorld ) );
	D3DXMatrixMultiply(&newWorld, &matTransform, &oldWorld);
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_WORLD, &newWorld ) );
	V_RETURN( DrawPrimitive(PrimitiveType) );
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_WORLD, &oldWorld ) );
	return S_OK;
}