//--------------------------------------------------------------------------------------
// File: Viewport.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Viewport class source
// Implementation of CViewport class
//
//--------------------------------------------------------------------------------------

#include "dxstdafx.h"

#pragma warning(disable : 4530)

#include <cmath>

#include "Viewport.h"

//--------------------------------------------------------------------------------------
// CViewport class
//--------------------------------------------------------------------------------------
HRESULT CViewport::SetViewport()
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;

	if (!m_bClippingRect)
		return E_FAIL;

	SetMatrices();

	V_RETURN( m_pd3dDevice->GetViewport(&m_oldArea) );

	V_RETURN( m_pd3dDevice->GetTransform( D3DTS_WORLD, &m_oldWorld ) );
	V_RETURN( m_pd3dDevice->GetTransform( D3DTS_VIEW, &m_oldView ) );
	V_RETURN( m_pd3dDevice->GetTransform( D3DTS_PROJECTION, &m_oldProjection ) );

	V_RETURN( m_pd3dDevice->SetViewport(&m_viewArea) );

	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_matrixWorld ) );
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_matrixView ) );
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matrixProjection ) );

	m_bViewportActivated = true;

	return S_OK;
}

HRESULT CViewport::ClearViewport(D3DCOLOR Color)
{
	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;

	if (m_bViewportActivated)
	{
		V_RETURN( m_pd3dDevice->Clear
			(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, Color, 1.0f, 0) );
		return S_OK;
	}
	else return E_FAIL;
}

HRESULT CViewport::UndoViewport()
{
	if (!m_bClippingRect)
		return E_FAIL;

	HRESULT hr;
	if (!m_pd3dDevice) return E_FAIL;

	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_WORLD, &m_oldWorld ) );
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_oldView ) );
	V_RETURN( m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_oldProjection ) );

	V_RETURN( m_pd3dDevice->SetViewport(&m_oldArea) );

	m_bViewportActivated = false;

	return S_OK;
}

void CViewport::SetClippingRect(int X, int Y, int Width, int Height)
{
	m_viewArea.X = X;
	m_viewArea.Y = Y;
	m_viewArea.Width = Width;
	m_viewArea.Height = Height;
	m_viewArea.MinZ = 0.0f;
	m_viewArea.MaxZ = 1.0f;
	m_bClippingRect = true;
	m_bUpdateMatrices = true;
}

void CViewport::SetToTarget(float Width, float Height)
{
	if (Width  < 1.0f) Width = 1.0f;
	if (Height < 1.0f) Height = 1.0f;
	m_fX = Width / 2;
	m_fY = Height / 2;

	if (Width > Height)
		m_fScale = 1.9f/(float)Width;
	else
		m_fScale = 1.9f/(float)Height;
	m_bUpdateMatrices = true;
}

void CViewport::MoveTo(float fX, float fY, float fElapsedTime)
{
	float fPart = (float)(std::exp(std::log(0.25) * fElapsedTime));
	m_fX = fX + fPart * (m_fX - fX);
	m_fY = fY + fPart * (m_fY - fY);
	m_bUpdateMatrices = true;
}

void CViewport::ScaleTo(float fScale, float fElapsedTime)
{
	float fPart = (float)(std::exp(std::log(0.25) * fElapsedTime));
	m_fScale = fScale + fPart * (m_fScale - fScale);
	m_bUpdateMatrices = true;
}

void CViewport::SetMatrices() const
{
	if (m_bUpdateMatrices)
	{
		D3DXMATRIX matrixTranslation;
		D3DXMATRIX matrixScaling;
		D3DXMATRIX matrixTmp;

		FLOAT Determinant;

		float width  = (float)m_viewArea.Width;
		float height = (float)m_viewArea.Height;
		float fx = 1.0f;
		float fy = 1.0f;

		if (width >= height)
			fx = height / width;
		else fy = width / height;

		D3DXMatrixTranslation(&matrixTranslation,
			-m_fX, -m_fY, 0.0f);
		D3DXMatrixScaling(&matrixScaling,
			m_fScale, m_fScale, 1.0f);

		D3DXMatrixIdentity(&m_matrixWorld);
		D3DXMatrixMultiply(&m_matrixView, &matrixTranslation, &matrixScaling);
		D3DXMatrixScaling(&m_matrixProjection, fx, fy, 1.0f);

		D3DXMatrixMultiply(&matrixTmp, &m_matrixWorld, &m_matrixView);
		D3DXMatrixMultiply(&m_matrixProduct, &matrixTmp, &m_matrixProjection);
		D3DXMatrixInverse (&m_matrixProductInverse, &Determinant, &m_matrixProduct);

		m_bUpdateMatrices = false;
	}
}

bool CViewport::IsInside(int globalX, int globalY) const
{
	return
		(globalX >= (int)m_viewArea.X && globalX < (int)m_viewArea.X + (int)m_viewArea.Width &&
		 globalY >= (int)m_viewArea.Y && globalY < (int)m_viewArea.Y + (int)m_viewArea.Height);
}

void CViewport::GetGlobalPoint(int & globalX, int & globalY, float localX, float localY) const
{
	D3DXVECTOR2 localVector(localX, localY);
	GetGlobalPoint(globalX, globalY, localVector);
}

void CViewport::GetGlobalPoint(int & globalX, int & globalY, const D3DXVECTOR2 & localVector) const
{
	SetMatrices();
	D3DXVECTOR2 Vector;
	D3DXVec2TransformCoord(&Vector, &localVector, &m_matrixProduct);
	globalX = m_viewArea.X + (int)((float)m_viewArea.Width  * (1.0f + Vector.x) / 2.0f);
	globalY = m_viewArea.Y + (int)((float)m_viewArea.Height * (1.0f - Vector.y) / 2.0f);
}

void CViewport::GetLocalPoint (int globalX, int globalY, float & localX, float & localY) const
{
	D3DXVECTOR2 localVector;
	GetLocalPoint(globalX, globalY, localVector);
	localX = localVector.x;
	localY = localVector.y;
}

void CViewport::GetLocalPoint (int globalX, int globalY, D3DXVECTOR2 & localVector) const
{
	SetMatrices();
	float fX = + 2*((float)globalX - (float)m_viewArea.X) / (float)m_viewArea.Width  - 1.0f;
	float fY = - 2*((float)globalY - (float)m_viewArea.Y) / (float)m_viewArea.Height + 1.0f;
	D3DXVECTOR2 globalVector(fX, fY);
	D3DXVECTOR2 Vector;
	D3DXVec2TransformCoord(&Vector, &globalVector, &m_matrixProductInverse);
	localVector.x = Vector.x;
	localVector.y = Vector.y;
}

HRESULT CViewport::OnCreatedDevice(IDirect3DDevice9 *pd3dDevice)
{
	m_pd3dDevice = pd3dDevice;
	return S_OK;
}

HRESULT CViewport::OnResetDevice()
{
	m_bClippingRect = false;
	return S_OK;
}

HRESULT CViewport::OnLostDevice()
{
	m_bClippingRect = false;
	return S_OK;
}

HRESULT CViewport::OnDestroyDevice()
{
	m_bClippingRect = false;
	return S_OK;
}