/*
    liban - base function to do usefull things related to simulation and scientific work
    Copyright (C) 1999-2005 Stephane Magnenat <nct@ysagoon.com>
    Copyright (c) 2004-2005 Antoine Beyeler <antoine.beyeler@epfl.ch>
    Copyright (C) 2005 Laboratory of Intelligent Systems, EPFL, Lausanne
    See AUTHORS for details

    This program is free software. You can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Formula.h"
#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <iostream>
#include <string>
#include <assert.h>
#include <algorithm>

using namespace std;
using namespace boost::spirit;
using namespace phoenix;
using namespace boost::lambda;

/*!	\file Formula.cpp
	\brief Implementation of Forumla, a formule parser, bytecompiler and interpreter
*/

namespace An
{
	
	//! Bytecode for variable, holds a pointer to variable
	struct VariableBytecodeToken : public Formula::BytecodeToken
	{
		const double *variable; //!< pointer to variable
		static double zero; //!< a double variable whose value is zero
		
		VariableBytecodeToken(Formula::VariableMap *_map, const std::string &name)
		{
			Formula::VariableMap::iterator it = _map->find(name);
			if (it == _map->end())
			{
				cerr << "Variable " << name << " doesn't exists, assigning zero" << std::endl;
				variable = &zero;
			}
			else
			{
				variable = it->second;
			}
		}
		
		virtual bool eval(Formula::ResultStack &stack)
		{
			stack.push(*variable);
			return true;
		}
	};
	
	double VariableBytecodeToken::zero = 0;
	
	//! Bytecode for function call, holds a pointer to functor
	struct FunctionBytecodeToken :  public Formula::BytecodeToken
	{
		Formula::UnaryFunctionPtr funcPtr; //!< pointer to functor
		
		FunctionBytecodeToken(Formula::FunctionMap *_funcMap, const std::string &name)
		{
			Formula::FunctionMap::iterator it = _funcMap->find(name);
			if (it == _funcMap->end())
			{
				cerr << "Function " << name << " doesn't exists, using identity function" << endl;
				funcPtr = &identity;
			}
			else
			{
				funcPtr = it->second;
			}
		}
		
		virtual bool eval(Formula::ResultStack &stack)
		{
			assert(stack.size() >= 1);
			
			double v = stack.top();
			stack.pop();
			stack.push(funcPtr(v));
			
			return true;
		}
		
		//! identity function
		static double identity(double v)
		{
			return v;
		}
	};
	
	//! Bytecode for constant, holds a value
	struct ConstantBytecodeToken : public Formula::BytecodeToken
	{
		const double value; //!< constant value
		
		ConstantBytecodeToken(double d) : value(d) { }
		
		virtual bool eval(Formula::ResultStack &stack)
		{
			stack.push(value);
			return true;
		}
	};
	
	//! Bytecode for unary functions, pop value from stack and push result of function. Templatised on the functor
	template<class Operation>
	struct UnaryFunctionBytecodeToken : public Formula::BytecodeToken
	{
		virtual bool eval(Formula::ResultStack &stack)
		{
			assert (stack.size() >= 1);
			
			double v = stack.top();
			stack.pop();
			stack.push(Operation()(v));
			
			return true;
		}
	};
	
	//! Negation functor
	struct UnaryNegationOperation
	{
		double operator()(double v) { return -v; }
	};
	//! Bytecode for negation
	typedef UnaryFunctionBytecodeToken<UnaryNegationOperation> UnaryNegationBytecodeToken;
	
	//! Bytecode for unary functions, pop two values from stack and push result of function. Templatised on the functor
	template<class Operation>
	struct BinaryFunctionBytecodeToken : public Formula::BytecodeToken
	{
		virtual bool eval(Formula::ResultStack &stack)
		{
			assert(stack.size() >= 2);
			
			double v1 = stack.top();
			stack.pop();
			double v0 = stack.top();
			stack.pop();
			stack.push(Operation()(v0, v1));
			
			return true;
		}
	};
	
	//! Addition functor
	struct AdditionOperator
	{
		double operator()(double v0, double v1) { return v0 + v1; }
	};
	//! Bytecode for addition
	typedef BinaryFunctionBytecodeToken<AdditionOperator> AdditionBytecodeToken;
	
	//! Substraction functor
	struct SubstractionOperator
	{
		double operator()(double v0, double v1) { return v0 - v1; }
	};
	//! Bytecode for substraction
	typedef BinaryFunctionBytecodeToken<SubstractionOperator> SubstractionBytecodeToken;
	
	//! Multiplication functor
	struct MultiplicationOperator
	{
		double operator()(double v0, double v1) { return v0 * v1; }
	};
	//! Bytecode for multiplication
	typedef BinaryFunctionBytecodeToken<MultiplicationOperator> MultiplicationBytecodeToken;
	
	//! Division functor
	struct DivisionOperator
	{
		double operator()(double v0, double v1) { return v0 / v1; }
	};
	//! Bytecode for division
	typedef BinaryFunctionBytecodeToken<DivisionOperator> DivisionBytecodeToken;
	
	//! Functor that add a new operator at the end of _bytecode when called
	template<class TokenType>
	struct PushOperator
	{
		PushOperator(Formula::Bytecode *_bytecode) { this->_bytecode = _bytecode; }
		
		void operator()(char const*, char const*) const { _bytecode->push_back(new TokenType); }
		
		Formula::Bytecode *_bytecode;
	};
	
	//! Functor that add a new constant at the end of _bytecode when called
	struct PushConstant
	{
		PushConstant(Formula::Bytecode *_bytecode) { this->_bytecode = _bytecode; }
		
		void operator()(const double value) const { _bytecode->push_back(new ConstantBytecodeToken(value)); }
		
		Formula::Bytecode *_bytecode;
	};
	
	//! Functor that add a new function call at the end of _bytecode when called
	struct PushFunction
	{
		PushFunction(Formula::Bytecode *_bytecode, Formula::FunctionMap *_funcMap) { this->_bytecode = _bytecode; this->_funcMap = _funcMap; }
		
		void operator()(const std::string &s) const
		{
			_bytecode->push_back(new FunctionBytecodeToken(_funcMap, s));
		}
		
		Formula::Bytecode *_bytecode;
		Formula::FunctionMap *_funcMap;
		std::string s;
	};
	
	//! Functor that add a new variable reading at the end of _bytecode when called
	struct PushVariable
	{
		PushVariable(Formula::Bytecode *_bytecode, Formula::VariableMap *_map) { this->_bytecode = _bytecode; this->_map = _map;}
		
		void operator()(const std::string &s) const
		{
			_bytecode->push_back(new VariableBytecodeToken(_map, s));
		}
		
		Formula::Bytecode *_bytecode;
		Formula::VariableMap *_map;
	};
	
	//! boost::spirit housekeeping
	struct StringClosure : boost::spirit::closure<StringClosure, std::string>
	{
		member1 val;
	};
	
	//! Language grammar using boost::spirit meta-programming features. You now enter highly magic area, prepare to forget everything you have believed in. Read it at your own risks, after all, this is in a cpp not h file.
	struct FormulaGrammar : public grammar<FormulaGrammar, StringClosure::context_t>
	{
		//! Initialisation of grammar, copy some reference for further use
		FormulaGrammar(Formula::Bytecode *_bytecode, Formula::VariableMap *_map, Formula::FunctionMap *_funcMap)
		{
			this->_bytecode = _bytecode;
			this->_map = _map;
			this->_funcMap = _funcMap;
		}
		
		//! EBNF grammar, read source and boost::spirit documentation and rethink your life
		template <typename ScannerT>
		struct definition
		{
			// rules for this grammar
			rule<ScannerT> real;
			rule<ScannerT, StringClosure::context_t> identifier;
			rule<ScannerT, StringClosure::context_t> function;
			rule<ScannerT> functionBuilder;
			rule<ScannerT> variable;
			rule<ScannerT> factor;
			rule<ScannerT> term;
			rule<ScannerT> expression;
			std::string n;
			
			// definition constructor
			definition(FormulaGrammar const& self)
			{
				real =
						real_p[PushConstant(self._bytecode)]
					;
					
				identifier =
						lexeme_d
						[
							(alpha_p >> *(alnum_p | '_'))
						][identifier.val = construct_<std::string>(arg1, arg2)]
					;
					
				function =
						identifier[function.val = arg1] >> '(' >> expression >> ')'
					;
					
				functionBuilder =
						function[PushFunction(self._bytecode, self._funcMap)]
					;
				
				variable =
						identifier[PushVariable(self._bytecode, self._map)]
					;
					
				factor =
					    real
					|   functionBuilder
					|   variable
					|   '(' >> expression >> ')'
					|   ('-' >> factor)[PushOperator<UnaryNegationBytecodeToken>(self._bytecode)]
					|   ('+' >> factor)
					;
			
				term =
					factor
					>> *(   ('*' >> factor)[PushOperator<MultiplicationBytecodeToken>(self._bytecode)]
						|   ('/' >> factor)[PushOperator<DivisionBytecodeToken>(self._bytecode)]
						)
						;
			
				expression =
					term
					>> *(   ('+' >> term)[PushOperator<AdditionBytecodeToken>(self._bytecode)]
						|   ('-' >> term)[PushOperator<SubstractionBytecodeToken>(self._bytecode)]
						)
						;
			}
			
			rule<ScannerT> const& start() const { return expression; }
		};
		
		Formula::Bytecode *_bytecode;
		Formula::VariableMap *_map;
		Formula::FunctionMap *_funcMap;
	};
	
	Formula::Formula(const std::string &formula, bool useStandardFunctions)
	{
		formulaText = formula;
		
		// put standard function
		if(useStandardFunctions)
		{
			functionMap["sin"] = &sin;
			functionMap["cos"] = &cos;
			functionMap["tan"] = &tan;
			functionMap["asin"] = &asin;
			functionMap["acos"] = &acos;
			functionMap["atan"] = &atan;
			functionMap["sinh"] = &sinh;
			functionMap["cosh"] = &cosh;
			functionMap["tanh"] = &tanh;
			functionMap["asinh"] = &asinh;
			functionMap["acosh"] = &acosh;
			functionMap["atanh"] = &atanh;
			functionMap["sqrt"] = &sqrt;
			functionMap["sqr"] = &sqr;
			functionMap["exp"] = &exp;
			functionMap["log"] = &log;
			functionMap["abs"] = &fabs;
			functionMap["sgn"] = &sgn;
		}
		
		parsed = false;
	}
	
	Formula::~Formula()
	{
		for (size_t i = 0; i < bytecode.size(); i++)
			delete bytecode[i];
	}
	
	void Formula::setVariable(const std::string &name, const double &value)
	{
		variableMap[name] = &value;
	}
	
	Formula::ParseResult Formula::parse(void)
	{
		// clean bytecode
		for (size_t i = 0; i < bytecode.size(); i++)
			delete bytecode[i];
		bytecode.resize(0);
		
		// parse
		FormulaGrammar formulaParser(&bytecode, &variableMap, &functionMap);
		parse_info<> info = boost::spirit::parse(formulaText.c_str(), formulaParser, space_p);
		
		ParseResult result;
		result.result = ParseResult::SUCCESS;
		
		/*std::cout << "Dumping formula tokens" << std::endl;
		for (size_t i=0; i<bytecode.size(); i++)
			std::cout << bytecode[i]->getType() << std::endl;*/
		
		// mark as parsed
		parsed = true;
		
		return result;
	}
	
	double Formula::eval(void)
	{
		ResultStack stack;
		size_t i = 0;
		bool result = true;
		
		// check if the formula was parsed
		if(!parsed)
			parse();
		
		// execute
		while ((i < bytecode.size()) && ((result = bytecode[i]->eval(stack)) == true))
			i++;
		
		if (result)
		{
			assert(stack.size() == 1);
			return stack.top();
		}
		return 0;
	}
}
