//--------------------------------------------------------------------------------------
// File: NeuralRobot.cpp
// Author: Peter Cerno <petercerno@gmail.com>
//
// Enki extension
// Implementation of Enki::NeuralRobot class
//
//--------------------------------------------------------------------------------------

#pragma warning(disable : 4530)
#pragma warning(disable : 4996)

#include "NeuralRobot.h"

namespace Enki
{
	NeuralRobotIRSensorModel neuralRobotIRSensorModel;
	SensorResponseFunctor *neuralRobotIRSensorModelPtr = &neuralRobotIRSensorModel;
		
	NeuralRobot::NeuralRobot():
		robotSpeed(1.0),
		wheelDist(1.0),
		leftEncoder(0.0),
		rightEncoder(0.0),
		leftSpeed(0.0),
		rightSpeed(0.0),
		oldAngle(0.0),
		remoteControl(0)
	{

	}

	NeuralRobot::~NeuralRobot()
	{
		Clear();
	}
	
	void NeuralRobot::Clear()
	{
		size_t i = 0;
		ClearRemoteControl();
		for (i = 0; i < infraredSensors.size(); ++i)
			delete infraredSensors[i];
		for (i = 0; i < cameras.size(); ++i)
			delete cameras[i];
		infraredSensors.resize(0);
		cameras.resize(0);
		robotSpeed = 1.0;
		wheelDist  = 1.0;
	}

	void NeuralRobot::ClearRemoteControl()
	{
		if (remoteControl)
		{
			remoteControl->setOwner(0);
			remoteControl = 0;
		}
	}

	void NeuralRobot::SetRemoteControl(RemoteControl * r)
	{
		ClearRemoteControl();
		remoteControl = r;
		remoteControl->setOwner(this);
	}

	void NeuralRobot::step(double dt) 
	{
		leftSpeed = 0.0;
		rightSpeed = 0.0;
		if (remoteControl)
		{
			remoteControl->Update();
			if (remoteControl->isUp())
			{
				leftSpeed += robotSpeed;
				rightSpeed += robotSpeed;
			}
			else if (remoteControl->isDown())
			{
				leftSpeed -= robotSpeed;
				rightSpeed -= robotSpeed;
			}
			if (remoteControl->isLeft())
				rightSpeed += robotSpeed;
			if (remoteControl->isRight())
				leftSpeed += robotSpeed;
		}
		// +/- 5% motor noise
		double realLeftSpeed  = leftSpeed  * (0.95 + random.getRange(0.1));
		double realRightSpeed = rightSpeed * (0.95 + random.getRange(0.1));
		
		// speeds, according to Prof. Siegwart class material
		double forwardSpeed = (realRightSpeed + realLeftSpeed) * 0.5;
		angSpeed += (realRightSpeed-realLeftSpeed) / wheelDist;
		speed = An::Vector(
					forwardSpeed * cos(angle + angSpeed * dt * 0.5),
					forwardSpeed * sin(angle + angSpeed * dt * 0.5));
		
		// PhysicalObject::step will actually move, and in next loop
		// care will be taken regarding collision. So we have to compute
		// the difference here.
		An::Vector posDiff = pos - oldPos;
		double norm = posDiff.norm();
		double travelAngle = posDiff.angle();
			
		// we let only the component of the norm that is in the direction
		// of the robot (the other component is the wheel sliding in the
		// perpendicular direction). we take the mean angle as direction.
		// angle and oldAngle are normalized, so we dont have problems here.
		double meanAngle = (angle + oldAngle) * 0.5;
		if(fabs(angle-oldAngle) > M_PI)
			meanAngle += M_PI;	// this is not normalized but we dont care
		norm = norm * cos(meanAngle - travelAngle);
		
		// Compute encoders
		double angleDiff = An::normalizeAngle(angle - oldAngle);
		leftEncoder = (norm - wheelDist * angleDiff * 0.5) / dt;
		rightEncoder = (norm + wheelDist * angleDiff * 0.5) / dt;
		
		// Save values for next step.
		oldPos = pos;
		oldAngle = angle;
		
		// handle physic
		PhysicalObject::step(dt);
	}
	
	void NeuralRobot::resetEncoders()
	{
		oldPos = pos;
		oldAngle = angle;
		leftEncoder = rightEncoder = 0.0;
	}
}