#ifndef __NEURALROBOT_H
#define __NEURALROBOT_H

//--------------------------------------------------------------------------------------
// File: NeuralRobot.h
// Author: Peter Cerno <petercerno@gmail.com>
//
// Enki extension
// Enki::NeuralRobot class subclassed from Enki::Robot
//
//--------------------------------------------------------------------------------------

#include <vector>

#include "../PhysicalEngine.h"
#include "../interactions/IRSensor.h"
#include "../interactions/CircularCam.h"

namespace Enki
{
	class NeuralRobot;

	//! Calculate the signal strength as a function of the distance.
	/*! The nearer we are, the higher the sensor activation.
		This model is very simple and not very good but sufficient for simple demonstration. */

	struct NeuralRobotIRSensorModel : public SensorResponseFunctor
	{
		virtual double operator()(double dist, const An::Color &)
		{
			dist -= 1.0;
			if(dist < 0.0)
				dist = 0.0;
			// add proportional noise; +/-3% and fit
			dist *= (0.97 + random.getRange(0.06));
			return 1100.0 * exp(-dist / 2.5);
		}
	};
	
	//! We use only one ray per sensor for the SimpleRobot.
	extern NeuralRobotIRSensorModel neuralRobotIRSensorModel;
	//! Pointer to sensor model for SimpleRobot, one value C array.
	extern SensorResponseFunctor *neuralRobotIRSensorModelPtr;

	//! Remote control
	class RemoteControl
	{
	protected:
		RemoteControl(const RemoteControl &);
		const RemoteControl & operator=(const RemoteControl &);
	public:
		RemoteControl():
		  owner(0) { }
		virtual ~RemoteControl() { }

		virtual bool isUp() = 0;
		virtual bool isDown() = 0;
		virtual bool isLeft() = 0;
		virtual bool isRight() = 0;

		virtual void Clear() = 0;
		virtual void Update() = 0;

		void setOwner(NeuralRobot * o) { owner = o; }
	protected:
		NeuralRobot * owner;
	};

	//! Enki::NeuralRobot class subclassed from Enki::Robot
	class NeuralRobot : public Robot
	{
	protected:
		NeuralRobot(const NeuralRobot &);
		const NeuralRobot & operator=(const NeuralRobot &);
	public:
		//! The infrared sensors
		std::vector<IRSensor*> infraredSensors;
		//! Linear cameras
		std::vector<CircularCam*> cameras;	
	public:
		double robotSpeed;
		//! The bot's wheel distance
		double wheelDist;
	protected:
		//! The bot's left and right wheel speed [1:100] cm/sec
		double leftSpeed, rightSpeed;
	public:
		//! The encoder for left and right wheel speed
		double leftEncoder, rightEncoder;
	protected:
		//! Save of last position to compute encoders.
		An::Vector oldPos;
		//! Save of last angle to compute encoders.
		double oldAngle;
	public:
		//! Create a NeuralRobot
		NeuralRobot();
		//! Destructor
		virtual ~NeuralRobot();

		void Clear();
		void ClearRemoteControl();

		void SetRemoteControl(RemoteControl * r);
		RemoteControl * GetRemoteControl() { return remoteControl; }

		//! Call NeuralRobot::controlStep and do all the calculations
		void step(double dt);
		
		//! Reset the encoder. Should be called when robot is moved manually.
		void resetEncoders();

	protected:
		RemoteControl * remoteControl;
	};
}

#endif // __NEURALROBOT_H