/*
    Enki - a fast 2D robot simulator
    Copyright (C) 1999-2005 Stephane Magnenat <nct@ysagoon.com>
    Copyright (C) 2004-2005 Markus Waibel <markus.waibel@epfl.ch>
    Copyright (C) 2005 Laboratory of Intelligent Systems, EPFL, Lausanne
    See AUTHORS for details

    This program is free software; the authors of any publication 
    arising from research using this software are asked to add the 
    following reference:
    Enki - a fast 2D robot simulator part of the Teem framework
    http://teem.epfl.ch
    Stephane Magnenat <stephane.magnenat@epfl.ch>,
    Markus Waibel <markus.waibel@epfl.ch>
    Laboratory of Intelligent Systems, EPFL, Lausanne.

    You can redistribute this program and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma warning(disable : 4530)
#pragma warning(disable : 4996)

#include "Alice.h"


// Implementation of the Alice robot

namespace Enki
{
	using namespace An;
	
	 // In our case, the sensormodelfunction is of the form a*exp(-b*x)-c*exp(-d*x)+e,
	 //   where e is a global offset, sometimes adjusted with respect to sensorOffsetNoise.
	
	//! The bounding surface of an Alice
	struct AliceBoundingSurface
	{
		//! The polygone containing the surface
		Polygone p;

		//! The constructor, fill p with (-1,-1) (1,-1) (1,1) (-1,1)
		AliceBoundingSurface()
		{
			// Basic
// 			p.push_back(Point(-1.25,-1));
// 			p.push_back(Point(1.25,-1));
// 			p.push_back(Point(1.25,1));
// 			p.push_back(Point(-1.25,1));
			
			// Snowplough
// 			p.push_back(Point(-1.25,-1));
// 			p.push_back(Point(1.05,-1));
// 			p.push_back(Point(1.26,0));
// 			p.push_back(Point(1.05,1));
// 			p.push_back(Point(-1.25,1));
			
			// Double Snowplough
// 			p.push_back(Point(-1.15,-1));
// 			p.push_back(Point(1.15,-1));
// 			p.push_back(Point(1.25,0));
// 			p.push_back(Point(1.15,1));
// 			p.push_back(Point(-1.15,1));
// 			p.push_back(Point(-1.25,0));
			
			// Cut-off Corners
// 			p.push_back(Point(-1.25,-1));
// 			p.push_back(Point(1.0,-1));
// 			p.push_back(Point(1.25,-.75));
// 			p.push_back(Point(1.25,.75));
// 			p.push_back(Point(1.0,1));
// 			p.push_back(Point(-1.25,1));

			// Double-quadruple cut-off corners
			p.push_back(Point(-1.0,-1));
			p.push_back(Point(1.0,-1));
			p.push_back(Point(1.2,-.7));
			p.push_back(Point(1.25,-.15));
			p.push_back(Point(1.25,.15));
			p.push_back(Point(1.2,.7));
			p.push_back(Point(1.0,1));
			p.push_back(Point(-1.0,1));
			p.push_back(Point(-1.1,0));
 		}
	} aliceBoundingSurface;


	//! Calculate the signal strength as a function of the distance.
	//! The nearer we are, the higher the sensor activation.

	struct AliceIRNormalSensorModel : public SensorResponseFunctor
	{
		double mult; //!< multiplier coefficient for response computation
		double shift; //!< shift coefficients for response computation

		//! Constructor, store coefficients for response computation
		AliceIRNormalSensorModel(double mult, double shift)
		{
			this->mult = mult;
			this->shift = shift;
		}

		virtual double operator()(double dist, const An::Color &)
		{
			// last term adjusts for half of overall offset so noise is +/-
			return shift + mult * (230.973 * exp (-1.49667*dist) - 108.396 * exp (-5.43925*dist) + 1.0 - 1.5);
		}
	};

	//! Calculate the signal strength as a function of the distance. 
	//! Same as AliceIRNormalSensorModel, but works at twice the distance.
	//! The nearer we are, the higher the sensor activation.

	struct AliceIRHighSensorModel : public SensorResponseFunctor
	{
		double mult; //!< multiplier coefficient for response computation
		double shift; //!< shift coefficients for response computation
		
		//! Constructor, store coefficients for response computation
		AliceIRHighSensorModel(double mult, double shift)
		{
			this->mult = mult;
			this->shift = shift;
		}

		virtual double operator()(double dist, const An::Color &)
		{
			// linear (shift) and multiplicative (mult) offset calculated at initialisation of sensormodel
			return shift + mult * (111.803 * exp (-0.2986*dist) - 97.4358 * exp (-2.96755*dist) + 2.45098 - 3.5);
		}
	};

	//! Response function which returns zero all the time.
	struct ZeroSensorModel : public SensorResponseFunctor
	{
		//! Return zero
		virtual double operator()(double, const An::Color &) { return 0; }
	};

	Alice::SensorModels::SensorModels()
	{
		left[0] = new AliceIRNormalSensorModel(0.345098, random.getRange(3.0));
		left[1] = new AliceIRNormalSensorModel(0.309804, random.getRange(3.0));
		left[2] = new AliceIRNormalSensorModel(0.345098, random.getRange(3.0));
		normal[0] = new AliceIRNormalSensorModel(0.345098, random.getRange(3.0));
		normal[1] = new AliceIRNormalSensorModel(0.309804, random.getRange(3.0));
		normal[2] = new AliceIRNormalSensorModel(0.345098, random.getRange(3.0));
		high[0] = new AliceIRHighSensorModel(0.333, random.getRange(6.0));
		high[1] = new AliceIRHighSensorModel(0.333, random.getRange(6.0));
		high[2] = new AliceIRHighSensorModel(0.333, random.getRange(6.0));
	}

	Alice::SensorModels::~SensorModels()
	{
		for (size_t i = 0; i<3; i++)
		{
			delete left[i];
			delete normal[i];
			delete high[i];
		}
	}

	Alice::Alice(unsigned capabilities) :
		sensorModels(),
	// IRSensor(Robot *owner, An::Vector pos, double height, double orientation, double range,
	//   double aperture, unsigned rayCount, double *rayCombinationKernel, SensorResponseFunction sensorModel);
	// the radius is the max range of the IR Sensor.

		left(Vector (0.788235, 0.944118),     1.2, 0.73030,   2.5,  0.77883,  3, sensorModels.normal),
		front(Vector (1.25, 0),        		1.2, 0,         2.5,  1.5291,   3, sensorModels.normal),
		right(Vector (0.788235, -0.944118),   1.2,-0.73030,   2.5,  0.77883,  3, sensorModels.normal),
 		back(Vector (-0.95, 0),        		1.2, M_PI,      2.5,  M_PI/6,   3, sensorModels.normal),
		// IR Sensor on new extension module with camera
		frontHighExt(Vector (0.9 , 0),    2.4, 0,         9.0, 1.285,  3, sensorModels.high),
		// back left sensor of the new ext module; global angle of 160 deg.
		backleft(Vector (-0.95, 1.05), 1.2, 8*M_PI/9, 2.5,  M_PI/6,  3, sensorModels.normal),
		backright(Vector(-0.95, -1.05),1.2, -8*M_PI/9,  2.5,  M_PI/6,  3, sensorModels.normal),
		// CircularCam(Robot *owner, An::Vector pos, double height = 0, double orientation,
		//   double fieldOfView, unsigned pixelCount);
		circcam(Vector (0, 0), 3, 0, M_PI/10, 6),
		comm(4.6)
	{
		left.setOwner(this);
		front.setOwner(this);
		right.setOwner(this);
		back.setOwner(this);
		frontHighExt.setOwner(this);
		backleft.setOwner(this);
		backright.setOwner(this);
		circcam.setOwner(this);
		comm.setOwner(this);
		mass = 10;
		height = 4;
		setBoundingSurface(aliceBoundingSurface.p);

		// set high static friction threshold to exclude Alices pushed by disks and 
		// delay Alice-Alice pushing (factor 2 slowdown; since collisions are symmetric, 
		// we don't know which Alice is pushing which one)
		staticFrictionThreshold = 1.0;
		collisionAngularFrictionFactor = 1.7;
		viscousFrictionTau = 0.5;
		viscousMomentFrictionTau = 0.0;
		
		if (capabilities & CAPABILITIY_BASIC_SENSORS)
		{
			addLocalInteraction(&left);
			addLocalInteraction(&front);
			addLocalInteraction(&right);
			addLocalInteraction(&back);
		}
		if (capabilities & CAPABILITIY_CAMERA)
		{
			addLocalInteraction(&circcam);
		}
		if (capabilities & CAPABILITIY_COMMUNICATION)
		{
			comm.addSensor(Point(-2.1,2.1), 1.6);
			comm.addSensor(Point(0,2.6), 1.6);
			comm.addSensor(Point(2.1,2.1), 1.6);
			comm.addSensor(Point(0,-2.6), 1.6);	
			addLocalInteraction(&comm);
		}
		if (capabilities & CAPABILITIY_EXT_HIGH_SENSOR)
		{
			addLocalInteraction(&frontHighExt);
		}
		if (capabilities & CAPABILITIY_EXT_BACK_LR_SENSORS)
		{
			addLocalInteraction(&backleft);
			addLocalInteraction(&backright);
		}

		leftSpeed = rightSpeed = 0;
	}

	void Alice::step(double dt) 
	{
		double realLeftSpeed, realRightSpeed;
		
		// rightSpeed/leftSpeed are the speeds as given by the NN output 
		// function,i.e. between -1 and 1. we do a speed transformation,
		// eliminating small Alice motor speeds, since the stepmotors 
		// do not function correctly for small velocities
		// currently, possible speeds are between 2.5 and 4cm/s
		if (fabs(leftSpeed) < 0.625)
			realLeftSpeed = 0;
		else
			// we apply an inversely proportional motor noise to the speeds between 
			// -4 and +4; the noise is of the order of +/- 5% for the maximum speed, 
			// for +/- 3cm/s it is +/- 10%
			realLeftSpeed = 4 * leftSpeed * (0.95 + random.getRange(0.05));

		if (fabs(rightSpeed) < 0.625)
			realRightSpeed = 0;
		else
			// same as above
			realRightSpeed = 4 * rightSpeed * (0.95 + random.getRange(0.05));

		// forward component
		double forwardSpeed = (realLeftSpeed+realRightSpeed) / 2;
		double wheelDist = 1.9;
		// Khepera code:
		speed = Vector(
					forwardSpeed * cos(angle + angSpeed * dt * 0.5),
					forwardSpeed * sin(angle + angSpeed * dt * 0.5));

		// angle
		angSpeed += (-realLeftSpeed+realRightSpeed) / wheelDist;

		// handle physics
		PhysicalObject::step(dt);
	}
}
